@@ -13,8 +13,7 @@
  * GNU General Public License for more details.
  *
  * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+ * along with this program; if not,  see <http://www.gnu.org/licenses>
  *
  * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  */
@@ -27,9 +26,10 @@
 #include "trace-local.h"
 #include "trace-view-store.h"
 #include "trace-view.h"
+#include "trace-gui.h"
 
 #include "cpu.h"
-#include "util.h"
+#include "event-utils.h"
 #include "list.h"
 
 #define DIALOG_WIDTH	400
@@ -58,11 +58,6 @@ int id_cmp(const void *a, const void *b)
 	return 0;
 }
 
-struct dialog_helper {
-	GtkWidget		*dialog;
-	gpointer		data;
-};
-
 /**
  * trace_array_add - allocate and add an int to an array.
  * @array: address of array to allocate
@@ -97,18 +92,12 @@ struct event_combo_info {
 	GtkWidget		*event_combo;
 	GtkWidget		*op_combo;
 	GtkWidget		*field_combo;
+	GtkWidget		*entry;
 };
 
-struct adv_event_filter_helper {
-	trace_adv_filter_cb_func	func;
-	GtkTreeView			*view;
-	GtkWidget			*entry;
-	struct event_combo_info		combo_info;
-	gpointer			data;
-};
-
-static GtkTreeModel *create_event_combo_model(struct pevent *pevent)
+static GtkTreeModel *create_event_combo_model(gpointer data)
 {
+	struct pevent *pevent = data;
 	GtkTreeStore *tree;
 	GtkTreeIter sys_iter;
 	GtkTreeIter iter;
@@ -143,7 +132,7 @@ static GtkTreeModel *create_event_combo_model(struct pevent *pevent)
 	return GTK_TREE_MODEL(tree);
 }
 
-static GtkTreeModel *create_op_combo_model(struct pevent *pevent)
+static GtkTreeModel *create_op_combo_model(gpointer data)
 {
 	GtkListStore *list;
 	GtkTreeIter iter;
@@ -165,8 +154,9 @@ static GtkTreeModel *create_op_combo_model(struct pevent *pevent)
 	return GTK_TREE_MODEL(list);
 }
 
-static GtkTreeModel *create_field_combo_model(struct pevent *pevent)
+static GtkTreeModel *create_field_combo_model(gpointer data)
 {
+	struct pevent *pevent = data;
 	GtkListStore *list;
 	GtkTreeIter iter;
 	struct event_format **events;
@@ -319,7 +309,6 @@ static void event_combo_changed(GtkComboBox *combo, gpointer data)
 static void insert_combo_text(struct event_combo_info *info,
 			      GtkComboBox *combo)
 {
-	struct adv_event_filter_helper *event_helper;
 	GtkTreeModel *model;
 	GtkTreeIter iter;
 	GtkWidget *entry;
@@ -337,8 +326,7 @@ static void insert_combo_text(struct event_combo_info *info,
 			   0, &text,
 			   -1);
 
-	event_helper = container_of(info, typeof(*event_helper), combo_info);
-	entry = event_helper->entry;
+	entry = info->entry;
 
 	pos = gtk_editable_get_position(GTK_EDITABLE(entry));
 	gtk_editable_insert_text(GTK_EDITABLE(entry), text, strlen(text), &pos);
@@ -375,12 +363,9 @@ static void field_insert_pressed(GtkButton *button,
 
 static GtkWidget *
 create_combo_box(struct event_combo_info *info, GtkWidget *hbox, const gchar *text,
-		 GtkTreeModel *(*combo_model_create)(struct pevent *pevent),
+		 GtkTreeModel *(*combo_model_create)(gpointer data),
 		 void (*insert_pressed)(GtkButton *button, gpointer data))
 {
-	GtkCellRenderer *renderer;
-	GtkTreeModel *model;
-	GtkWidget *label;
 	GtkWidget *hbox2;
 	GtkWidget *combo;
 	GtkWidget *button;
@@ -389,33 +374,7 @@ create_combo_box(struct event_combo_info *info, GtkWidget *hbox, const gchar *te
 	gtk_box_pack_start(GTK_BOX(hbox), hbox2, TRUE, TRUE, 0);
 	gtk_widget_show(hbox2);
 
-	label = gtk_label_new(text);
-	gtk_box_pack_start(GTK_BOX(hbox2), label, FALSE, FALSE, 0);
-	gtk_widget_show(label);
-
-	/* --- Set up the selection combo box --- */
-
-	model = combo_model_create(info->pevent);
-
-	renderer = gtk_cell_renderer_text_new();
-
-	combo = gtk_combo_box_new_with_model(model);
-	gtk_box_pack_start(GTK_BOX(hbox2), combo, FALSE, FALSE, 0);
-	gtk_widget_show(combo);
-
-	/* Free model with combobox */
-	g_object_unref(model);
-
-	gtk_cell_layout_pack_start(GTK_CELL_LAYOUT(combo),
-				   renderer,
-				   TRUE);
-	gtk_cell_layout_set_attributes(GTK_CELL_LAYOUT(combo),
-				       renderer,
-				       "text", 0,
-				       NULL);
-
-	gtk_combo_box_set_active(GTK_COMBO_BOX(combo), 0);
-
+	combo = trace_create_combo_box(hbox, text, combo_model_create, info->pevent);
 
 	/* --- add insert button --- */
 
@@ -506,48 +465,16 @@ static gint *get_event_ids(GtkTreeView *treeview)
 	return ids;
 }
 
-/* Callback for the clicked signal of the advanced filter button */
-static void
-adv_filter_dialog_response (gpointer data, gint response_id)
-{
-	struct dialog_helper *helper = data;
-	struct adv_event_filter_helper *event_helper = helper->data;
-	const gchar *text;
-	gint *event_ids;
-
-	switch (response_id) {
-	case GTK_RESPONSE_ACCEPT:
-		text = gtk_entry_get_text(GTK_ENTRY(event_helper->entry));
-		event_ids = get_event_ids(event_helper->view);
-		event_helper->func(TRUE, text, event_ids, event_helper->data);
-		free(event_ids);
-		break;
-	case GTK_RESPONSE_REJECT:
-		event_helper->func(FALSE, NULL, NULL, event_helper->data);
-		break;
-	default:
-		break;
-	};
-
-	gtk_widget_destroy(GTK_WIDGET(helper->dialog));
-
-	g_free(event_helper);
-	g_free(helper);
-}
-
 static GtkTreeModel *
-create_tree_filter_model(struct tracecmd_input *handle,
+create_tree_filter_model(struct pevent *pevent,
 		       struct event_filter *event_filter)
 {
 	GtkTreeStore *treestore;
 	GtkTreeIter iter_events;
-	struct pevent *pevent;
 	struct event_format **events;
 	char *str;
 	gint i;
 
-	pevent = tracecmd_get_pevent(handle);
-
 	treestore = gtk_tree_store_new(NUM_ADV_FILTER_COLS, G_TYPE_BOOLEAN,
 				       G_TYPE_STRING, G_TYPE_STRING,
 				       G_TYPE_INT);
@@ -625,7 +552,7 @@ static void adv_filter_cursor_changed(GtkTreeView *treeview, gpointer data)
 }
 
 static GtkWidget *
-create_adv_filter_view(struct tracecmd_input *handle,
+create_adv_filter_view(struct pevent *pevent,
 		       struct event_filter *event_filter)
 {
 	GtkTreeViewColumn *col;
@@ -677,7 +604,7 @@ create_adv_filter_view(struct tracecmd_input *handle,
 	gtk_tree_view_column_add_attribute(col, renderer, "text", ADV_COL_FILTER);
 
 
-	model = create_tree_filter_model(handle, event_filter);
+	model = create_tree_filter_model(pevent, event_filter);
 
 	gtk_tree_view_set_model(GTK_TREE_VIEW(view), model);
 
@@ -706,9 +633,8 @@ void trace_adv_filter_dialog(struct tracecmd_input *handle,
 			       trace_adv_filter_cb_func func,
 			       gpointer data)
 {
+	struct event_combo_info combo_info;
 	struct pevent *pevent;
-	struct dialog_helper *helper;
-	struct adv_event_filter_helper *event_helper;
 	GtkWidget *dialog;
 	GtkWidget *hbox;
 	GtkWidget *label;
@@ -716,12 +642,16 @@ void trace_adv_filter_dialog(struct tracecmd_input *handle,
 	GtkWidget *scrollwin;
 	GtkWidget *view;
 	GtkWidget *event_box;
+	const gchar *text;
+	gint *event_ids;
+	int result;
 
 	if (!handle)
 		return;
 
-	helper = g_malloc(sizeof(*helper));
-	g_assert(helper);
+	pevent = tracecmd_get_pevent(handle);
+	if (!pevent)
+		return;
 
 	/* --- Make dialog window --- */
 
@@ -734,26 +664,11 @@ void trace_adv_filter_dialog(struct tracecmd_input *handle,
 					     GTK_RESPONSE_REJECT,
 					     NULL);
 
-	event_helper = g_new0(typeof(*event_helper), 1);
-	g_assert(event_helper);
-
-	helper->dialog = dialog;
-	helper->data = event_helper;
-
-	event_helper->func = func;
-	event_helper->data = data;
-
-	/* We can attach the Quit menu item to our exit function */
-	g_signal_connect_swapped (dialog, "response",
-				  G_CALLBACK (adv_filter_dialog_response),
-				  (gpointer) helper);
-
 	scrollwin = gtk_scrolled_window_new(NULL, NULL);
 	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrollwin),
 				       GTK_POLICY_AUTOMATIC,
 				       GTK_POLICY_AUTOMATIC);
-	view = create_adv_filter_view(handle, event_filter);
-	event_helper->view = GTK_TREE_VIEW(view);
+	view = create_adv_filter_view(pevent, event_filter);
 	gtk_container_add(GTK_CONTAINER(scrollwin), view);
 	gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox), scrollwin, TRUE, TRUE, 0);
 
@@ -767,11 +682,9 @@ void trace_adv_filter_dialog(struct tracecmd_input *handle,
 	gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox), label, TRUE, FALSE, 0);
 	gtk_widget_show(label);
 
-	pevent = tracecmd_get_pevent(handle);
+	combo_info.pevent = pevent;
 
-	event_helper->combo_info.pevent = pevent;
-
-	event_box = event_info_box(&event_helper->combo_info);
+	event_box = event_info_box(&combo_info);
 	gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox), event_box, FALSE, FALSE, 0);
 	gtk_widget_show(event_box);
 
@@ -787,23 +700,33 @@ void trace_adv_filter_dialog(struct tracecmd_input *handle,
 	gtk_box_pack_start(GTK_BOX(hbox), entry, TRUE, TRUE, 0);
 	gtk_widget_show(entry);
 
-	event_helper->entry = entry;
+	combo_info.entry = entry;
 
 	gtk_widget_set_size_request(GTK_WIDGET(dialog),
 				    TEXT_DIALOG_WIDTH, TEXT_DIALOG_HEIGHT);
 
 	gtk_widget_show_all(dialog);
+
+	result = gtk_dialog_run(GTK_DIALOG(dialog));
+	switch (result) {
+	case GTK_RESPONSE_ACCEPT:
+		text = gtk_entry_get_text(GTK_ENTRY(entry));
+		event_ids = get_event_ids(GTK_TREE_VIEW(view));
+		func(TRUE, text, event_ids, data);
+		free(event_ids);
+		break;
+	case GTK_RESPONSE_REJECT:
+		func(FALSE, NULL, NULL, data);
+		break;
+	default:
+		break;
+	};
+
+	gtk_widget_destroy(dialog);
 }
 
 /* --- task list dialog --- */
 
-struct task_helper {
-	trace_task_cb_func		func;
-	GtkTreeView			*view;
-	gboolean			start;
-	gpointer			data;
-};
-
 enum {
 	TASK_COL_SELECT,
 	TASK_COL_PID,
@@ -850,43 +773,13 @@ static void get_tasks(GtkTreeView *treeview,
 	*non_selected = non_pids;
 }
 
-/* Callback for the clicked signal of the task filter button */
-static void
-task_dialog_response (gpointer data, gint response_id)
-{
-	struct dialog_helper *helper = data;
-	struct task_helper *event_helper = helper->data;
-	gint *selected;
-	gint *non_select;
-
-	switch (response_id) {
-	case GTK_RESPONSE_ACCEPT:
-		get_tasks(event_helper->view, &selected, &non_select);
-		event_helper->func(TRUE, selected, non_select, event_helper->data);
-		free(selected);
-		free(non_select);
-		break;
-	case GTK_RESPONSE_REJECT:
-		event_helper->func(FALSE, NULL, NULL, event_helper->data);
-		break;
-	default:
-		break;
-	};
-
-	gtk_widget_destroy(GTK_WIDGET(helper->dialog));
-
-	g_free(event_helper);
-	g_free(helper);
-}
-
 static GtkTreeModel *
-create_task_model(struct tracecmd_input *handle,
+create_task_model(struct pevent *pevent,
 		  gint *tasks,
 		  gint *selected)
 {
 	GtkTreeStore *treestore;
 	GtkTreeIter iter;
-	struct pevent *pevent;
 	const char *comm;
 	gboolean select;
 	gint *ret;
@@ -897,8 +790,6 @@ create_task_model(struct tracecmd_input *handle,
 	if (!tasks)
 		return NULL;
 
-	pevent = tracecmd_get_pevent(handle);
-
 	treestore = gtk_tree_store_new(NUM_TASK_COLS, G_TYPE_BOOLEAN,
 				       G_TYPE_INT, G_TYPE_STRING);
 
@@ -960,14 +851,14 @@ create_task_model(struct tracecmd_input *handle,
 
 static void task_cursor_changed(gpointer data, GtkTreeView *treeview)
 {
-	struct task_helper *event_helper = data;
+	gboolean *start = data;
 	GtkTreeModel *model;
 	GtkTreePath *path;
 	GtkTreeIter iter;
 	gboolean active;
 
-	if (!event_helper->start) {
-		event_helper->start = TRUE;
+	if (!*start) {
+		*start = TRUE;
 		return;
 	}
 
@@ -997,9 +888,9 @@ static void task_cursor_changed(gpointer data, GtkTreeView *treeview)
 }
 
 static GtkWidget *
-create_task_view(struct tracecmd_input *handle,
+create_task_view(struct pevent *pevent,
 		 gint *tasks, gint *selected,
-		 struct task_helper *event_helper)
+		 gboolean *start)
 {
 	GtkTreeViewColumn *col;
 	GtkCellRenderer *renderer;
@@ -1050,7 +941,7 @@ create_task_view(struct tracecmd_input *handle,
 	gtk_tree_view_column_add_attribute(col, renderer, "text", TASK_COL_COMM);
 
 
-	model = create_task_model(handle, tasks, selected);
+	model = create_task_model(pevent, tasks, selected);
 
 	gtk_tree_view_set_model(GTK_TREE_VIEW(view), model);
 
@@ -1061,7 +952,7 @@ create_task_view(struct tracecmd_input *handle,
 
 	g_signal_connect_swapped (view, "cursor-changed",
 				  G_CALLBACK (task_cursor_changed),
-				  (gpointer) event_helper);
+				  (gpointer)start);
 
 	return view;
 }
@@ -1080,14 +971,20 @@ void trace_task_dialog(struct tracecmd_input *handle,
 		       trace_task_cb_func func,
 		       gpointer data)
 {
-	struct dialog_helper *helper;
-	struct task_helper *event_helper;
+	struct pevent *pevent;
 	GtkWidget *dialog;
 	GtkWidget *scrollwin;
 	GtkWidget *view;
+	gboolean start = FALSE;
+	gint *non_select;
+	int result;
 
-	helper = g_malloc(sizeof(*helper));
-	g_assert(helper);
+	if (!handle)
+		return;
+
+	pevent = tracecmd_get_pevent(handle);
+	if (!pevent)
+		return;
 
 	/* --- Make dialog window --- */
 
@@ -1100,27 +997,11 @@ void trace_task_dialog(struct tracecmd_input *handle,
 					     GTK_RESPONSE_REJECT,
 					     NULL);
 
-	event_helper = g_new0(typeof(*event_helper), 1);
-	g_assert(event_helper);
-
-	helper->dialog = dialog;
-	helper->data = event_helper;
-
-	event_helper->func = func;
-	event_helper->data = data;
-	event_helper->start = FALSE;
-
-	/* We can attach the Quit menu item to our exit function */
-	g_signal_connect_swapped (dialog, "response",
-				  G_CALLBACK (task_dialog_response),
-				  (gpointer) helper);
-
 	scrollwin = gtk_scrolled_window_new(NULL, NULL);
 	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrollwin),
 				       GTK_POLICY_AUTOMATIC,
 				       GTK_POLICY_AUTOMATIC);
-	view = create_task_view(handle, tasks, selected, event_helper);
-	event_helper->view = GTK_TREE_VIEW(view);
+	view = create_task_view(pevent, tasks, selected, &start);
 	gtk_container_add(GTK_CONTAINER(scrollwin), view);
 	gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox), scrollwin, TRUE, TRUE, 0);
 
@@ -1128,6 +1009,25 @@ void trace_task_dialog(struct tracecmd_input *handle,
 				    DIALOG_WIDTH, DIALOG_HEIGHT);
 
 	gtk_widget_show_all(dialog);
+
+	selected = NULL;
+
+	result = gtk_dialog_run(GTK_DIALOG(dialog));
+	switch (result) {
+	case GTK_RESPONSE_ACCEPT:
+		get_tasks(GTK_TREE_VIEW(view), &selected, &non_select);
+		func(TRUE, selected, non_select, data);
+		free(selected);
+		free(non_select);
+		break;
+	case GTK_RESPONSE_REJECT:
+		func(FALSE, NULL, NULL, data);
+		break;
+	default:
+		break;
+	};
+
+	gtk_widget_destroy(dialog);
 }
 
 enum {
@@ -1170,14 +1070,13 @@ gboolean event_is_enabled(gint *events, gint events_size, gint event)
 }
 
 static GtkTreeModel *
-create_tree_event_model(struct tracecmd_input *handle,
+create_tree_event_model(struct pevent *pevent,
 			struct event_filter *filter,
 			gboolean all_events, gchar **systems_set,
 			gint *event_ids_set)
 {
 	GtkTreeStore *treestore;
 	GtkTreeIter iter_all, iter_sys, iter_events;
-	struct pevent *pevent;
 	struct event_format **events;
 	struct event_format *event;
 	char *last_system = NULL;
@@ -1189,8 +1088,6 @@ create_tree_event_model(struct tracecmd_input *handle,
 	gint event_ids_size;
 	gint i;
 
-	pevent = tracecmd_get_pevent(handle);
-
 	treestore = gtk_tree_store_new(NUM_EVENT_COLS, G_TYPE_STRING,
 				       G_TYPE_BOOLEAN, G_TYPE_BOOLEAN,
 				       G_TYPE_BOOLEAN, G_TYPE_INT);
@@ -1260,6 +1157,7 @@ create_tree_event_model(struct tracecmd_input *handle,
 				   -1);
 	}
 
+	/* These are copies of the sets, using glib allocation */
 	g_free(systems);
 	g_free(event_ids);
 
@@ -1457,11 +1355,54 @@ static void expand_rows(GtkTreeView *tree, GtkTreeModel *model,
 	}
 }
 
-static GtkWidget *
-create_event_list_view(struct tracecmd_input *handle,
-		       struct event_filter *filter,
-		       gboolean all_events, gchar **systems,
-		       gint *events)
+/**
+ * trace_update_event_view - update the events of an existing event view
+ * @event_view: event view to update
+ * @pevent: The parse event descriptor
+ * @filter: Event filter to determine what events have advanced filters
+ *          May be NULL.
+ * @all_events: True if all should be selected,
+ * @systems: Array of system names of systems that should be selected.
+ * @events: Array of event ids of events that should be selecetd.
+ */
+int trace_update_event_view(GtkWidget *event_view,
+			    struct pevent *pevent,
+			    struct event_filter *filter,
+			    gboolean all_events,
+			    gchar **systems, gint *events)
+{
+	GtkTreeView *view = GTK_TREE_VIEW(event_view);
+	GtkTreeModel *model;
+
+	model = create_tree_event_model(pevent, filter,
+					all_events, systems, events);
+	if (!model)
+		return -1;
+
+	gtk_tree_view_set_model(view, model);
+	g_object_unref(model);
+
+	expand_rows(view, model, all_events, systems, events);
+
+	return 0;
+}
+
+/**
+ * trace_create_event_list_view - create a list view of events in pevent
+ * @pevent: The parse event descriptor
+ * @filter: Event filter to determine what events have advanced filters
+ *          May be NULL.
+ * @all_events: True if all should be selected,
+ * @systems: Array of system names of systems that should be selected.
+ * @events: Array of event ids of events that should be selected.
+ *
+ * Returns a tree view widget of the events.
+ */
+GtkWidget *
+trace_create_event_list_view(struct pevent *pevent,
+			     struct event_filter *filter,
+			     gboolean all_events, gchar **systems,
+			     gint *events)
 {
 	GtkTreeViewColumn *col;
 	GtkCellRenderer *renderer;
@@ -1491,7 +1432,7 @@ create_event_list_view(struct tracecmd_input *handle,
 	gtk_tree_view_column_add_attribute(col, renderer, "text", COL_EVENT);
 	gtk_tree_view_column_add_attribute(col, renderer, "sensitive", COL_NORMAL);
 
-	model = create_tree_event_model(handle, filter, all_events, systems, events);
+	model = create_tree_event_model(pevent, filter, all_events, systems, events);
 
 	gtk_tree_view_set_model(GTK_TREE_VIEW(view), model);
 
@@ -1509,36 +1450,6 @@ create_event_list_view(struct tracecmd_input *handle,
 	return view;
 }
 
-static gchar **add_system(gchar **systems, gint size, gchar *system)
-{
-	if (!systems) {
-		systems = g_new0(gchar *, 2);
-		size = 0;
-	} else {
-		systems = g_realloc(systems,
-				    sizeof(*systems) * (size + 2));
-	}
-	systems[size] = g_strdup(system);
-	systems[size+1] = NULL;
-
-	return systems;
-}
-
-static gint *add_event(gint *events, gint size, gint event)
-{
-	if (!events) {
-		events = g_new0(gint, 2);
-		size = 0;
-	} else {
-		events = g_realloc(events,
-				   sizeof(*events) * (size + 2));
-	}
-	events[size] = event;
-	events[size+1] = -1;
-
-	return events;
-}
-
 static gint update_events(GtkTreeModel *model,
 			  GtkTreeIter *parent,
 			  gint **events, gint size)
@@ -1558,7 +1469,7 @@ static gint update_events(GtkTreeModel *model,
 				   -1);
 
 		if (active)
-			*events = add_event(*events, size++, id);
+			*events = tracecmd_add_id(*events, id, size++);
 
 		if (!gtk_tree_model_iter_next(model, &event))
 			break;
@@ -1567,19 +1478,19 @@ static gint update_events(GtkTreeModel *model,
 	return size;
 }
 
-static gint update_system_events(GtkTreeModel *model,
+static void update_system_events(GtkTreeModel *model,
 				 GtkTreeIter *parent,
 				 gchar ***systems,
-				 gint size,
-				 gint **events,
-				 gint *events_size)
+				 gint **events)
 {
 	GtkTreeIter sys;
 	gboolean active;
 	gchar *system;
+	gint sys_size = 0;
+	gint event_size = 0;
 
 	if (!gtk_tree_model_iter_children(model, &sys, parent))
-		return size;
+		return;
 
 	for (;;) {
 
@@ -1589,100 +1500,83 @@ static gint update_system_events(GtkTreeModel *model,
 				   -1);
 
 		if (active)
-			*systems = add_system(*systems, size++, system);
+			*systems = tracecmd_add_list(*systems, system, sys_size++);
 		else
-			*events_size = update_events(model, &sys, events, *events_size);
+			event_size = update_events(model, &sys, events, event_size);
 
 		g_free(system);
 
 		if (!gtk_tree_model_iter_next(model, &sys))
 			break;
 	}
-
-	return size;
 }
 
-static void accept_events(struct event_filter_helper *event_helper)
+/**
+ * trace_extract_event_list_view - extract selected events from event view
+ * @event_view: the event list view widget to extract from
+ * @all_events: Set if all events are set (@systems and @events are left as is)
+ * @systems: Set to the selected systems (if @all_events is FALSE)
+ * @events: Set to the selected events (not in selected systems and if @all_events is FALSE)
+ *
+ * Returns 0 on success, -1 otherwise.
+ */
+gint trace_extract_event_list_view(GtkWidget *event_view,
+				   gboolean *all_events,
+				   gchar ***systems,
+				   gint **events)
 {
-	GtkTreeView *view = event_helper->view;
+	GtkTreeView *view = GTK_TREE_VIEW(event_view);
 	GtkTreeModel *model;
 	GtkTreeIter iter;
-	gboolean active;
-	gchar **systems = NULL;
-	gint *events = NULL;
-	gint events_size = 0;
-	gint systems_size = 0;
-	gint i;
 
 	model = gtk_tree_view_get_model(view);
 	if (!model)
-		return;
+		return -1;
 
 	if (!gtk_tree_model_get_iter_first(model, &iter))
-		return;
+		return -1;
 
 	gtk_tree_model_get(model, &iter,
-			   COL_ACTIVE, &active,
+			   COL_ACTIVE, all_events,
 			   -1);
 
-	if (!active)
-		update_system_events(model, &iter,
-				     &systems, systems_size,
-				     &events, &events_size);
+	if (!*all_events)
+		update_system_events(model, &iter, systems, events);
 
-	event_helper->func(TRUE, active, systems, events,
-			   event_helper->data);
-
-	if (systems) {
-		for (i = 0; systems[i]; i++)
-			g_free(systems[i]);
-
-		g_free(systems);
-	}
-	g_free(events);
+	return 0;
 }
 
-/* Callback for the clicked signal of the Events filter button */
-static void
-event_dialog_response (gpointer data, gint response_id)
+static void accept_events(GtkWidget *view,
+			  trace_filter_event_cb_func func, gpointer data)
 {
-	struct dialog_helper *helper = data;
-	struct event_filter_helper *event_helper = helper->data;
+	gboolean all_events;
+	gchar **systems = NULL;
+	gint *events = NULL;
 
-	switch (response_id) {
-	case GTK_RESPONSE_ACCEPT:
-		printf("accept!\n");
-		accept_events(event_helper);
-		break;
-	case GTK_RESPONSE_REJECT:
-		printf("reject!\n");
-		event_helper->func(FALSE, FALSE, NULL, NULL,
-				   event_helper->data);
-		break;
-	default:
-		break;
-	};
+	if (trace_extract_event_list_view(view, &all_events,
+					  &systems, &events) < 0) {
+		/* Failed to extract ? */
+		func(FALSE, FALSE, NULL, NULL, data);
+		return;
+	}
 
-	gtk_widget_destroy(GTK_WIDGET(helper->dialog));
+	func(TRUE, all_events, systems, events, data);
 
-	g_free(event_helper);
-	g_free(helper);
+	tracecmd_free_list(systems);
+	free(events);
 }
 
-static void filter_event_dialog(struct tracecmd_input *handle,
+static void filter_event_dialog(struct pevent *pevent,
 				struct event_filter *filter,
 				gboolean all_events,
 				gchar **systems, gint *events,
 				trace_filter_event_cb_func func,
 				gpointer data)
 {
-	struct dialog_helper *helper;
-	struct event_filter_helper *event_helper;
 	GtkWidget *dialog;
 	GtkWidget *scrollwin;
 	GtkWidget *view;
-
-	helper = g_malloc(sizeof(*helper));
+	int result;
 
 	/* --- Make dialog window --- */
 
@@ -1695,26 +1589,11 @@ static void filter_event_dialog(struct tracecmd_input *handle,
 					     GTK_RESPONSE_REJECT,
 					     NULL);
 
-	event_helper = g_new0(typeof(*event_helper), 1);
-	g_assert(event_helper);
-
-	helper->dialog = dialog;
-	helper->data = event_helper;
-
-	event_helper->func = func;
-	event_helper->data = data;
-
-	/* We can attach the Quit menu item to our exit function */
-	g_signal_connect_swapped (dialog, "response",
-				  G_CALLBACK (event_dialog_response),
-				  (gpointer) helper);
-
 	scrollwin = gtk_scrolled_window_new(NULL, NULL);
 	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrollwin),
 				       GTK_POLICY_AUTOMATIC,
 				       GTK_POLICY_AUTOMATIC);
-	view = create_event_list_view(handle, filter, all_events, systems, events);
-	event_helper->view = GTK_TREE_VIEW(view);
+	view = trace_create_event_list_view(pevent, filter, all_events, systems, events);
 
 	gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox), scrollwin, TRUE, TRUE, 0);
 	gtk_container_add(GTK_CONTAINER(scrollwin), view);
@@ -1723,6 +1602,20 @@ static void filter_event_dialog(struct tracecmd_input *handle,
 				    DIALOG_WIDTH, DIALOG_HEIGHT);
 
 	gtk_widget_show_all(dialog);
+
+	result = gtk_dialog_run(GTK_DIALOG(dialog));
+	switch (result) {
+	case GTK_RESPONSE_ACCEPT:
+		accept_events(view, func, data);
+		break;
+	case GTK_RESPONSE_REJECT:
+		func(FALSE, FALSE, NULL, NULL, data);
+		break;
+	default:
+		break;
+	};
+
+	gtk_widget_destroy(dialog);
 }
 
 /**
@@ -1742,10 +1635,33 @@ void trace_filter_event_dialog(struct tracecmd_input *handle,
 			       trace_filter_event_cb_func func,
 			       gpointer data)
 {
-	filter_event_dialog(handle, NULL, all_events, systems,
+	struct pevent *pevent;
+
+	if (!handle)
+		return;
+
+	pevent = tracecmd_get_pevent(handle);
+	if (!pevent)
+		return;
+
+	filter_event_dialog(pevent, NULL, all_events, systems,
 			    events, func, data);
 }
 
+void trace_filter_pevent_dialog(struct pevent *pevent,
+				gboolean all_events,
+				gchar **systems, gint *events,
+				trace_filter_event_cb_func func,
+				gpointer data)
+{
+	if (!pevent)
+		return;
+
+	filter_event_dialog(pevent, NULL, all_events, systems,
+			    events, func, data);
+}
+
+
 /**
  * trace_filter_event_filter_dialog - make dialog with event listing
  * @handle: the handle to the tracecmd data file
@@ -1762,15 +1678,20 @@ void trace_filter_event_filter_dialog(struct tracecmd_input *handle,
 				      trace_filter_event_cb_func func,
 				      gpointer data)
 {
+	struct pevent *pevent;
 	gchar **systems;
 	gint *event_ids;
 
 	if (!handle)
 		return;
 
+	pevent = tracecmd_get_pevent(handle);
+	if (!pevent)
+		return;
+
 	trace_filter_convert_filter_to_names(filter, &systems, &event_ids);
 
-	filter_event_dialog(handle, filter, all_events, systems,
+	filter_event_dialog(pevent, filter, all_events, systems,
 			    event_ids, func, data);
 }
 
@@ -1779,8 +1700,6 @@ struct cpu_filter_helper {
 	guint64				*cpu_mask;
 	GtkWidget			**buttons;
 	int				cpus;
-	trace_filter_cpu_cb_func	func;
-	gpointer			data;
 };
 
 static void destroy_cpu_helper(struct cpu_filter_helper *cpu_helper)
@@ -1790,40 +1709,6 @@ static void destroy_cpu_helper(struct cpu_filter_helper *cpu_helper)
 	g_free(cpu_helper);
 }
 
-/* Callback for the clicked signal of the CPUS filter button */
-static void
-cpu_dialog_response (gpointer data, gint response_id)
-{
-	struct dialog_helper *helper = data;
-	struct cpu_filter_helper *cpu_helper = helper->data;
-	guint64 *cpu_mask = NULL;
-
-	switch (response_id) {
-	case GTK_RESPONSE_ACCEPT:
-
-		if (!cpu_helper->allcpus) {
-			cpu_mask = cpu_helper->cpu_mask;
-			cpu_helper->cpu_mask = NULL;
-		}
-
-		cpu_helper->func(TRUE, cpu_helper->allcpus, cpu_mask, cpu_helper->data);
-		break;
-
-	case GTK_RESPONSE_REJECT:
-		cpu_helper->func(FALSE, FALSE, NULL, cpu_helper->data);
-		break;
-	default:
-		break;
-	};
-
-	g_free(cpu_mask);
-
-	gtk_widget_destroy(GTK_WIDGET(helper->dialog));
-
-	destroy_cpu_helper(helper->data);
-	g_free(helper);
-}
-
 #define CPU_ALL_CPUS_STR "All CPUs"
 
 void cpu_toggle(gpointer data, GtkWidget *widget)
@@ -1838,29 +1723,36 @@ void cpu_toggle(gpointer data, GtkWidget *widget)
 
 	if (strcmp(label, CPU_ALL_CPUS_STR) == 0) {
 		cpu_helper->allcpus = active;
-		if (active) {
-			/* enable all toggles */
-			for (cpu = 0; cpu < cpu_helper->cpus; cpu++)
-				gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(cpu_helper->buttons[cpu]),
-							     TRUE);
-		}
+		/* enable/disable all toggles */
+		for (cpu = 0; cpu < cpu_helper->cpus; cpu++)
+			gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(cpu_helper->buttons[cpu]),
+						     active ? TRUE : FALSE);
 		return;
 	}
 
 	/* Get the CPU # from the label. Pass "CPU " */
 	cpu = atoi(label + 4);
-	if (active) {
+	if (active)
 		cpu_set(cpu_helper->cpu_mask, cpu);
-		return;
-	}
+	else
+		cpu_clear(cpu_helper->cpu_mask, cpu);
 
-	cpu_clear(cpu_helper->cpu_mask, cpu);
-
-	if (!cpu_helper->allcpus)
-		return;
+	cpu_helper->allcpus = cpu_allset(cpu_helper->cpu_mask,
+					 cpu_helper->cpus);
+	/*
+	 * Deactivate sending 'toggled' signal for "All CPUs"
+	 * while we adjust its state
+	 */
+	g_signal_handlers_block_by_func(GTK_TOGGLE_BUTTON(cpu_helper->buttons[cpu_helper->cpus]),
+				        G_CALLBACK (cpu_toggle),
+				        (gpointer) cpu_helper);
 
 	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(cpu_helper->buttons[cpu_helper->cpus]),
-				     FALSE);
+				     cpu_helper->allcpus);
+
+	g_signal_handlers_unblock_by_func(GTK_TOGGLE_BUTTON(cpu_helper->buttons[cpu_helper->cpus]),
+					  G_CALLBACK (cpu_toggle),
+					  (gpointer) cpu_helper);
 }
 
 /**
@@ -1875,8 +1767,8 @@ void cpu_toggle(gpointer data, GtkWidget *widget)
 void trace_filter_cpu_dialog(gboolean all_cpus, guint64 *cpus_selected, gint cpus,
 			     trace_filter_cpu_cb_func func, gpointer data)
 {
-	struct dialog_helper *helper;
 	struct cpu_filter_helper *cpu_helper;
+	guint64 *cpu_mask = NULL;
 	GtkWidget *dialog;
 	GtkWidget *scrollwin;
 	GtkWidget *viewport;
@@ -1888,15 +1780,11 @@ void trace_filter_cpu_dialog(gboolean all_cpus, guint64 *cpus_selected, gint cpu
 	gint width, height;
 	gint allset;
 	gint cpu;
-
-	helper = g_malloc(sizeof(*helper));
-	g_assert(helper != NULL);
+	int result;
 
 	cpu_helper = g_new0(typeof(*cpu_helper), 1);
 	g_assert(cpu_helper != NULL);
 
-	helper->data = cpu_helper;
-
 	/* --- Make dialog window --- */
 
 	dialog = gtk_dialog_new_with_buttons("Filter CPUS",
@@ -1908,19 +1796,10 @@ void trace_filter_cpu_dialog(gboolean all_cpus, guint64 *cpus_selected, gint cpu
 					     GTK_RESPONSE_REJECT,
 					     NULL);
 
-	helper->dialog = dialog;
-
 	cpu_helper->cpus = cpus;
 	cpu_helper->buttons = g_new0(GtkWidget *, cpus + 1);
 	g_assert(cpu_helper->buttons);
 
-	cpu_helper->func = func;
-	cpu_helper->data = data;
-
-	g_signal_connect_swapped (dialog, "response",
-				  G_CALLBACK (cpu_dialog_response),
-				  (gpointer) helper);
-
 	scrollwin = gtk_scrolled_window_new(NULL, NULL);
 	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrollwin),
 				       GTK_POLICY_AUTOMATIC,
@@ -2012,38 +1891,31 @@ void trace_filter_cpu_dialog(gboolean all_cpus, guint64 *cpus_selected, gint cpu
 				    width, height);
 
 	gtk_widget_show_all(dialog);
-}
 
-static void add_system_str(gchar ***systems, char *system, int count)
-{
-	if (!systems)
-		return;
+	result = gtk_dialog_run(GTK_DIALOG(dialog));
+	switch (result) {
+	case GTK_RESPONSE_ACCEPT:
 
-	if (!count)
-		*systems = malloc_or_die(sizeof(char *) * 2);
-	else
-		*systems = realloc(*systems, sizeof(char *) * (count + 2));
-	if (!*systems)
-		die("Can't allocate systems");
+		if (!cpu_helper->allcpus) {
+			cpu_mask = cpu_helper->cpu_mask;
+			cpu_helper->cpu_mask = NULL;
+		}
 
-	(*systems)[count] = system;
-	(*systems)[count+1] = NULL;
-}
+		func(TRUE, cpu_helper->allcpus, cpu_mask, data);
+		break;
 
-static void add_event_int(gint **events, gint event, int count)
-{
-	if (!events)
-		return;
+	case GTK_RESPONSE_REJECT:
+		func(FALSE, FALSE, NULL, data);
+		break;
+	default:
+		break;
+	};
 
-	if (!count)
-		*events = malloc_or_die(sizeof(gint) * 2);
-	else
-		*events = realloc(*events, sizeof(gint) * (count + 2));
-	if (!*events)
-		die("Can't allocate events");
+	g_free(cpu_mask);
 
-	(*events)[count] = event;
-	(*events)[count+1] = -1;
+	gtk_widget_destroy(dialog);
+
+	destroy_cpu_helper(cpu_helper);
 }
 
 /* -- Helper functions -- */
@@ -2087,7 +1959,7 @@ void trace_filter_convert_filter_to_names(struct event_filter *filter,
 		if (systems && last_system &&
 		    strcmp(last_system, events[i]->system) != 0) {
 			if (all_selected)
-				add_system_str(systems, last_system, sys_count++);
+				*systems = tracecmd_add_list(*systems, last_system, sys_count++);
 			start_sys = i;
 			all_selected = 1;
 		}
@@ -2095,11 +1967,11 @@ void trace_filter_convert_filter_to_names(struct event_filter *filter,
 		if (pevent_filter_event_has_trivial(filter, event->id,
 						    FILTER_TRIVIAL_TRUE)) {
 			if (!all_selected || !systems)
-				add_event_int(event_ids, event->id, event_count++);
+				*event_ids = tracecmd_add_id(*event_ids, event->id, event_count++);
 		} else {
 			if (all_selected && event_ids) {
 				for (x = start_sys; x < i; x++) {
-					add_event_int(event_ids,
+					*event_ids = tracecmd_add_id(*event_ids,
 						      events[x]->id, event_count++);
 				}
 			}
@@ -2107,13 +1979,13 @@ void trace_filter_convert_filter_to_names(struct event_filter *filter,
 
 			/* If this event is filtered, still add it */
 			if (pevent_event_filtered(filter, event->id))
-				add_event_int(event_ids, event->id, event_count++);
+				*event_ids = tracecmd_add_id(*event_ids, event->id, event_count++);
 		}
 		last_system = event->system;
 	}
 
 	if (systems && last_system && all_selected)
-		add_system_str(systems, last_system, sys_count++);
+		*systems = tracecmd_add_list(*systems, last_system, sys_count++);
 }
 
 /**
@@ -2158,3 +2030,236 @@ void trace_filter_convert_char_to_filter(struct event_filter *filter,
 
 	pevent_filter_free(copy);
 }
+
+int trace_filter_save_events(struct tracecmd_xml_handle *handle,
+			     struct event_filter *filter)
+{
+	struct event_format *event;
+	char **systems;
+	gint *event_ids;
+	char *str;
+	int i;
+
+	trace_filter_convert_filter_to_names(filter, &systems,
+					     &event_ids);
+
+	for (i = 0; systems && systems[i]; i++)
+		tracecmd_xml_write_element(handle, "System", systems[i]);
+
+	for (i = 0; event_ids && event_ids[i] > 0; i++) {
+		str = pevent_filter_make_string(filter, event_ids[i]);
+		if (!str)
+			continue;
+
+		event = pevent_find_event(filter->pevent, event_ids[i]);
+		if (event) {
+
+			/* skip not filtered items */
+			if (strcmp(str, "FALSE") == 0) {
+				free(str);
+				continue;
+			}
+
+			tracecmd_xml_start_sub_system(handle, "Event");
+			tracecmd_xml_write_element(handle, "System", event->system);
+			tracecmd_xml_write_element(handle, "Name", event->name);
+			/* If this is has an advanced filter, include that too */
+			if (strcmp(str, "TRUE") != 0) {
+				tracecmd_xml_write_element(handle, "Advanced",
+							   str);
+			}
+			tracecmd_xml_end_sub_system(handle);
+		}
+		free(str);
+	}
+
+	return 0;
+}
+
+int trace_filter_save_tasks(struct tracecmd_xml_handle *handle,
+			    struct filter_task *filter)
+{
+	char buffer[100];
+	int *pids;
+	int i;
+
+	pids = filter_task_pids(filter);
+	if (!pids)
+		return -1;
+
+	for (i = 0; pids[i] >= 0; i++) {
+		snprintf(buffer, 100, "%d", pids[i]);
+		tracecmd_xml_write_element(handle, "Task", buffer);
+	}
+
+	free(pids);
+
+	return 0;
+}
+
+int trace_filter_load_events(struct event_filter *event_filter,
+			     struct tracecmd_xml_handle *handle,
+			     struct tracecmd_xml_system_node *node)
+{
+	struct tracecmd_xml_system_node *child;
+	const char *name;
+	const char *system;
+	const char *event;
+	const char *value;
+	char *buffer;
+
+	while (node) {
+		name = tracecmd_xml_node_type(node);
+
+		if (strcmp(name, "System") == 0) {
+			system = tracecmd_xml_node_value(handle, node);
+			pevent_filter_add_filter_str(event_filter,
+						     system, NULL);
+		} else if (strcmp(name, "Event") == 0) {
+			system = NULL;
+			event = NULL;
+			value = NULL;
+			child = tracecmd_xml_node_child(node);
+			if (!child)
+				return -1;
+			do {
+				name = tracecmd_xml_node_type(child);
+				if (strcmp(name, "System") == 0)
+					system = tracecmd_xml_node_value(handle, child);
+				else if (strcmp(name, "Name") == 0)
+					event = tracecmd_xml_node_value(handle, child);
+				else if (strcmp(name, "Advanced") == 0)
+					value = tracecmd_xml_node_value(handle, child);
+				child = tracecmd_xml_node_next(child);
+			} while (child);
+
+			if (event || system) {
+				if (event && system) {
+					if (value) {
+						buffer = malloc_or_die(strlen(event) +
+								       strlen(system) +
+								       strlen(value) + 3);
+						sprintf(buffer, "%s/%s:%s",
+							system, event, value);
+					} else {
+						buffer = malloc_or_die(strlen(event) +
+								       strlen(system) + 2);
+						sprintf(buffer, "%s/%s",
+							system, event);
+					}
+				} else {
+					if (!event)
+						event = system;
+					if (value) {
+						buffer = malloc_or_die(strlen(event) +
+								       strlen(value) + 2);
+						sprintf(buffer, "%s:%s",
+							event, value);
+					} else {
+						buffer = malloc_or_die(strlen(event) + 1);
+						sprintf(buffer, "%s", event);
+					}
+				}
+				pevent_filter_add_filter_str(event_filter,
+							     buffer, NULL);
+				free(buffer);
+			}
+		}
+
+		node = tracecmd_xml_node_next(node);
+	}
+
+	return 0;
+}
+
+int trace_filter_load_task_filter(struct filter_task *filter,
+				  struct tracecmd_xml_handle *handle,
+				  struct tracecmd_xml_system_node *node)
+{
+	const char *name;
+	const char *task;
+	int pid;
+
+	if (!filter)
+		return 0;
+
+	node = tracecmd_xml_node_child(node);
+
+	while (node) {
+		name = tracecmd_xml_node_type(node);
+
+		if (strcmp(name, "Task") == 0) {
+			task = tracecmd_xml_node_value(handle, node);
+			pid = atoi(task);
+			if (!filter_task_find_pid(filter, pid))
+				filter_task_add_pid(filter, pid);
+		}
+		node = tracecmd_xml_node_next(node);
+	}
+
+	return 0;
+}
+
+int trace_filter_load_filters(struct tracecmd_xml_handle *handle,
+			      const char *system_name,
+			      struct filter_task *task_filter,
+			      struct filter_task *hide_tasks)
+{
+	struct tracecmd_xml_system *system;
+	struct tracecmd_xml_system_node *syschild;
+	const char *name;
+
+	system = tracecmd_xml_find_system(handle, system_name);
+	if (!system)
+		return -1;
+
+
+	syschild = tracecmd_xml_system_node(system);
+	if (!syschild)
+		goto out_free_sys;
+
+	do {
+		name = tracecmd_xml_node_type(syschild);
+
+		if (strcmp(name, "TaskFilter") == 0)
+			trace_filter_load_task_filter(task_filter, handle, syschild);
+
+		else if (strcmp(name, "HideTasks") == 0)
+			trace_filter_load_task_filter(hide_tasks, handle, syschild);
+
+		syschild = tracecmd_xml_node_next(syschild);
+	} while (syschild);
+
+	tracecmd_xml_free_system(system);
+
+	return 0;
+
+ out_free_sys:
+	tracecmd_xml_free_system(system);
+	return -1;
+}
+
+int trace_filter_save_filters(struct tracecmd_xml_handle *handle,
+			      const char *system_name,
+			      struct filter_task *task_filter,
+			      struct filter_task *hide_tasks)
+{
+
+	tracecmd_xml_start_system(handle, system_name);
+
+	if (task_filter && filter_task_count(task_filter)) {
+		tracecmd_xml_start_sub_system(handle, "TaskFilter");
+		trace_filter_save_tasks(handle, task_filter);
+		tracecmd_xml_end_sub_system(handle);
+	}
+
+	if (hide_tasks && filter_task_count(hide_tasks)) {
+		tracecmd_xml_start_sub_system(handle, "HideTasks");
+		trace_filter_save_tasks(handle, hide_tasks);
+		tracecmd_xml_end_sub_system(handle);
+	}
+
+	tracecmd_xml_end_system(handle);
+
+	return 0;
+}
