@@ -13,8 +13,7 @@
  * GNU General Public License for more details.
  *
  * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+ * along with this program; if not,  see <http://www.gnu.org/licenses>
  *
  * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  */
@@ -31,8 +30,9 @@
 #include "trace-graph.h"
 #include "trace-hash.h"
 #include "trace-filter.h"
+#include "trace-gui.h"
 
-#include "util.h"
+#include "event-utils.h"
 
 #define DEBUG_LEVEL	0
 #if DEBUG_LEVEL > 0
@@ -268,7 +268,7 @@ static void graph_filter_task_clear(struct graph_info *ginfo)
 	ginfo->filter_enabled = 0;
 }
 
-gboolean trace_graph_filter_on_event(struct graph_info *ginfo, struct record *record)
+gboolean trace_graph_filter_on_event(struct graph_info *ginfo, struct pevent_record *record)
 {
 	int ret;
 
@@ -326,6 +326,7 @@ static void update_label_time(GtkWidget *label, gint64 time)
 	trace_seq_printf(&s, "%s%lu.%06lu", min, sec, usec);
 
 	gtk_label_set_text(GTK_LABEL(label), s.buffer);
+	trace_seq_destroy(&s);
 }
 
 static void update_cursor(struct graph_info *ginfo)
@@ -379,7 +380,7 @@ static void draw_cursor(struct graph_info *ginfo)
 	x = convert_time_to_x(ginfo, ginfo->cursor);
 
 	gdk_draw_line(ginfo->draw->window, ginfo->draw->style->mid_gc[3],
-		      x, 0, x, ginfo->draw->allocation.width);
+		      x, 0, x, ginfo->draw->allocation.height);
 }
 
 static void draw_marka(struct graph_info *ginfo)
@@ -391,7 +392,7 @@ static void draw_marka(struct graph_info *ginfo)
 
 	x = convert_time_to_x(ginfo, ginfo->marka_time);
 	gdk_draw_line(ginfo->draw->window, green,
-		      x, 0, x, ginfo->draw->allocation.width);
+		      x, 0, x, ginfo->draw->allocation.height);
 }
 
 static void draw_markb(struct graph_info *ginfo)
@@ -403,7 +404,7 @@ static void draw_markb(struct graph_info *ginfo)
 
 	x = convert_time_to_x(ginfo, ginfo->markb_time);
 	gdk_draw_line(ginfo->draw->window, red,
-		      x, 0, x, ginfo->draw->allocation.width);
+		      x, 0, x, ginfo->draw->allocation.height);
 }
 
 static void update_with_backend(struct graph_info *ginfo,
@@ -433,7 +434,7 @@ static void
 draw_line(GtkWidget *widget, gdouble x, struct graph_info *ginfo)
 {
 	gdk_draw_line(widget->window, widget->style->black_gc,
-		      x, 0, x, widget->allocation.width);
+		      x, 0, x, widget->allocation.height);
 }
 
 static void clear_line(struct graph_info *ginfo, gint x)
@@ -548,6 +549,44 @@ void trace_graph_clear_tasks(struct graph_info *ginfo)
 		redraw_graph(ginfo);
 }
 
+void trace_graph_update_filters(struct graph_info *ginfo,
+				struct filter_task *task_filter,
+				struct filter_task *hide_tasks)
+{
+	/* Make sure the filter passed in is not the filter we use */
+	if (task_filter != ginfo->task_filter) {
+		filter_task_hash_free(ginfo->task_filter);
+		ginfo->task_filter = filter_task_hash_copy(task_filter);
+	}
+
+	if (hide_tasks != ginfo->hide_tasks) {
+		filter_task_hash_free(ginfo->hide_tasks);
+		ginfo->hide_tasks = filter_task_hash_copy(hide_tasks);
+	}
+
+	if (ginfo->callbacks && ginfo->callbacks->filter)
+		ginfo->callbacks->filter(ginfo, ginfo->task_filter,
+					 ginfo->hide_tasks);
+
+	if (ginfo->filter_enabled)
+		redraw_graph(ginfo);
+
+	if (filter_task_count(ginfo->task_filter) ||
+	    filter_task_count(ginfo->hide_tasks))
+		ginfo->filter_available = 1;
+	else {
+		ginfo->filter_enabled = 0;
+		ginfo->filter_available = 0;
+	}
+
+}
+
+void trace_graph_refresh_filters(struct graph_info *ginfo)
+{
+	trace_graph_update_filters(ginfo, ginfo->task_filter,
+				   ginfo->hide_tasks);
+}
+
 static void
 filter_clear_tasks_clicked (gpointer data)
 {
@@ -613,7 +652,7 @@ do_pop_up(GtkWidget *widget, GdkEventButton *event, gpointer data)
 	static GtkWidget *menu_filter_clear_tasks;
 	static GtkWidget *menu_plot_task;
 	static GtkWidget *menu_remove_plot;
-	struct record *record = NULL;
+	struct pevent_record *record = NULL;
 	struct graph_plot *plot;
 	const char *comm;
 	guint64 time;
@@ -718,7 +757,7 @@ do_pop_up(GtkWidget *widget, GdkEventButton *event, gpointer data)
 		g_assert(text);
 
 		if (trace_graph_filter_task_find_pid(ginfo, pid))
-			snprintf(text, len, "Remove %s-%d to filter", comm, pid);
+			snprintf(text, len, "Remove %s-%d from filter", comm, pid);
 		else
 			snprintf(text, len, "Add %s-%d to filter", comm, pid);
 
@@ -728,9 +767,9 @@ do_pop_up(GtkWidget *widget, GdkEventButton *event, gpointer data)
 					text);
 
 		if (trace_graph_hide_task_find_pid(ginfo, pid))
-			snprintf(text, len, "Show %s-%d to filter", comm, pid);
+			snprintf(text, len, "Show %s-%d", comm, pid);
 		else
-			snprintf(text, len, "Hide %s-%d to filter", comm, pid);
+			snprintf(text, len, "Hide %s-%d", comm, pid);
 
 		gtk_menu_item_set_label(GTK_MENU_ITEM(menu_filter_hide_task),
 					text);
@@ -752,7 +791,7 @@ do_pop_up(GtkWidget *widget, GdkEventButton *event, gpointer data)
 		gtk_widget_set_sensitive(menu_filter_add_task, FALSE);
 
 		gtk_menu_item_set_label(GTK_MENU_ITEM(menu_filter_hide_task),
-					"Hide task to filter");
+					"Hide task");
 		gtk_widget_set_sensitive(menu_filter_hide_task, FALSE);
 
 		gtk_widget_hide(menu_plot_task);
@@ -766,7 +805,25 @@ do_pop_up(GtkWidget *widget, GdkEventButton *event, gpointer data)
 	return TRUE;
 }
 
-static void button_press(struct graph_info *ginfo, gint x, guint state)
+static void draw_info_box(struct graph_info *ginfo, const gchar *buffer,
+			  gint x, gint y);
+
+static void stop_zoom_tip(struct graph_info *ginfo)
+{
+	clear_info_box(ginfo);
+}
+
+static void show_zoom_tip(struct graph_info *ginfo, gint x, gint y)
+{
+	clear_info_box(ginfo);
+
+	draw_info_box(ginfo,
+		      "Click and hold left mouse and drag right to zoom in\n"
+		      "Click and hold left mouse and drag left to zoom out",
+		      x, y);
+}
+
+static void button_press(struct graph_info *ginfo, gint x, gint y, guint state)
 {
 	ginfo->press_x = x;
 	ginfo->last_x = 0;
@@ -785,8 +842,10 @@ static void button_press(struct graph_info *ginfo, gint x, guint state)
 			clear_line(ginfo, convert_time_to_x(ginfo, ginfo->marka_time));
 			update_marka(ginfo, x);
 		}
-	} else
+	} else {
 		ginfo->zoom = TRUE;
+		show_zoom_tip(ginfo, x, y);
+	}
 
 	return;
 }
@@ -807,6 +866,7 @@ button_press_event(GtkWidget *widget, GdkEventButton *event, gpointer data)
 
 	/* check for double click */
 	if (event->type == GDK_2BUTTON_PRESS) {
+		stop_zoom_tip(ginfo);
 		if (ginfo->line_active) {
 			ginfo->line_active = FALSE;
 			clear_line(ginfo, ginfo->last_x);
@@ -826,7 +886,7 @@ button_press_event(GtkWidget *widget, GdkEventButton *event, gpointer data)
 		return TRUE;
 	}
 
-	button_press(ginfo, event->x, event->state);
+	button_press(ginfo, event->x, event->y, event->state);
 
 	return TRUE;
 }
@@ -839,6 +899,9 @@ static void motion_plot(struct graph_info *ginfo, gint x, gint y)
 {
 	struct graph_plot *plot;
 
+	if (ginfo->zoom)
+		stop_zoom_tip(ginfo);
+
 	if (!ginfo->curr_pixmap)
 		return;
 
@@ -876,7 +939,7 @@ info_button_press_event(GtkWidget *widget, GdkEventButton *event, gpointer data)
 	if (event->type == GDK_2BUTTON_PRESS)
 		return FALSE;
 
-	button_press(ginfo, gtk_adjustment_get_value(ginfo->hadj), event->state);
+	button_press(ginfo, gtk_adjustment_get_value(ginfo->hadj), event->y, event->state);
 
 	return FALSE;
 }
@@ -940,7 +1003,7 @@ info_button_release_event(GtkWidget *widget, GdkEventMotion *event, gpointer dat
 #define PLOT_BOARDER 5
 
 int trace_graph_check_sched_wakeup(struct graph_info *ginfo,
-				   struct record *record,
+				   struct pevent_record *record,
 				   gint *pid)
 {
 	struct event_format *event;
@@ -953,7 +1016,7 @@ int trace_graph_check_sched_wakeup(struct graph_info *ginfo,
 		found = FALSE;
 
 		event = pevent_find_event_by_name(ginfo->pevent,
-						  "sched", "sched_wakeup");
+						  NULL, "sched_wakeup");
 		if (event) {
 			found = TRUE;
 			ginfo->event_wakeup_id = event->id;
@@ -963,7 +1026,7 @@ int trace_graph_check_sched_wakeup(struct graph_info *ginfo,
 
 
 		event = pevent_find_event_by_name(ginfo->pevent,
-						  "sched", "sched_wakeup_new");
+						  NULL, "sched_wakeup_new");
 		if (event) {
 			found = TRUE;
 			ginfo->event_wakeup_new_id = event->id;
@@ -978,9 +1041,11 @@ int trace_graph_check_sched_wakeup(struct graph_info *ginfo,
 
 	if (id == ginfo->event_wakeup_id) {
 		/* We only want those that actually woke up the task */
-		pevent_read_number_field(ginfo->wakeup_success_field, record->data, &val);
-		if (!val)
-			return 0;
+		if (ginfo->wakeup_success_field) {
+			pevent_read_number_field(ginfo->wakeup_success_field, record->data, &val);
+			if (!val)
+				return 0;
+		}
 		pevent_read_number_field(ginfo->wakeup_pid_field, record->data, &val);
 		if (pid)
 			*pid = val;
@@ -989,9 +1054,11 @@ int trace_graph_check_sched_wakeup(struct graph_info *ginfo,
 
 	if (id == ginfo->event_wakeup_new_id) {
 		/* We only want those that actually woke up the task */
-		pevent_read_number_field(ginfo->wakeup_new_success_field, record->data, &val);
-		if (!val)
-			return 0;
+		if (ginfo->wakeup_new_success_field) {
+			pevent_read_number_field(ginfo->wakeup_new_success_field, record->data, &val);
+			if (!val)
+				return 0;
+		}
 		pevent_read_number_field(ginfo->wakeup_new_pid_field, record->data, &val);
 		if (pid)
 			*pid = val;
@@ -1002,7 +1069,7 @@ int trace_graph_check_sched_wakeup(struct graph_info *ginfo,
 }
 
 int trace_graph_check_sched_switch(struct graph_info *ginfo,
-				   struct record *record,
+				   struct pevent_record *record,
 				   gint *pid, const char **comm)
 {
 	unsigned long long val;
@@ -1019,7 +1086,7 @@ int trace_graph_check_sched_switch(struct graph_info *ginfo,
 
 	if (ginfo->event_sched_switch_id < 0) {
 		event = pevent_find_event_by_name(ginfo->pevent,
-						  "sched", "sched_switch");
+						  NULL, "sched_switch");
 		if (!event)
 			return 0;
 
@@ -1049,8 +1116,10 @@ int trace_graph_check_sched_switch(struct graph_info *ginfo,
 
 	if (id == ginfo->ftrace_sched_switch_id) {
 		pevent_read_number_field(ginfo->ftrace_pid_field, record->data, &val);
-		if (comm)
+		if (comm && ginfo->ftrace_comm_field)
 			*comm = record->data + ginfo->ftrace_comm_field->offset;
+		else
+			comm = NULL;
 		if (pid)
 			*pid = val;
 		goto out;
@@ -1065,7 +1134,7 @@ int trace_graph_check_sched_switch(struct graph_info *ginfo,
 		 */
 		if (!pevent_pid_is_registered(ginfo->pevent, *pid))
 			pevent_register_comm(ginfo->pevent,
-					     strdup(*comm), *pid);
+					     *comm, *pid);
 	}
 
 	return ret;
@@ -1161,6 +1230,7 @@ static void draw_plot_info(struct graph_info *ginfo, struct graph_plot *plot,
 
 	if (!trace_graph_plot_display_info(ginfo, plot, &s, time)) {
 		/* Just display the current time */
+		trace_seq_destroy(&s);
 		trace_seq_init(&s);
 		trace_seq_printf(&s, "%lu.%06lu", sec, usec);
 	}
@@ -1168,6 +1238,7 @@ static void draw_plot_info(struct graph_info *ginfo, struct graph_plot *plot,
 	trace_seq_putc(&s, 0);
 
 	draw_info_box(ginfo, s.buffer, x, y);
+	trace_seq_destroy(&s);
 }
 
 static void draw_latency(struct graph_info *ginfo, gint x, gint y)
@@ -1197,6 +1268,7 @@ static void draw_latency(struct graph_info *ginfo, gint x, gint y)
 	trace_seq_printf(&s, "Diff: %s%ld.%06lu secs", neg ? "-":"", sec, usec);
 
 	draw_info_box(ginfo, s.buffer, x, y);
+	trace_seq_destroy(&s);
 }
 
 static gboolean
@@ -1500,7 +1572,8 @@ static void button_release(struct graph_info *ginfo, gint x)
 		ginfo->show_marka = TRUE;
 		ginfo->show_markb = TRUE;
 		update_markb(ginfo, x);
-	}
+	} else
+		stop_zoom_tip(ginfo);
 
 	clear_line(ginfo, ginfo->last_x);
 	clear_line(ginfo, ginfo->press_x);
@@ -1597,13 +1670,17 @@ static gint draw_event_label(struct graph_info *ginfo, gint i,
 	 */
 	ret = trace_graph_plot_display_last_event(ginfo, plot, &s,
 						  convert_x_to_time(ginfo, p2-1));
-	if (!ret)
+	if (!ret) {
+		trace_seq_destroy(&s);
 		return p2;
+	}
 
 	layout = gtk_widget_create_pango_layout(ginfo->draw, s.buffer);
 	pango_layout_set_font_description(layout, font);
 	pango_layout_get_pixel_size(layout, &text_width, &text_height);
 
+	trace_seq_destroy(&s);
+
 	/* Lets see if we can print this info */
 	if (p2 < text_width)
 		start = 1;
@@ -1666,7 +1743,7 @@ static void draw_plot_box(struct graph_info *ginfo, int i,
 }
 
 static void draw_plot(struct graph_info *ginfo, struct graph_plot *plot,
-		      struct record *record)
+		      struct pevent_record *record)
 {
 	static PangoFontDescription *font;
 	PangoLayout *layout;
@@ -1735,7 +1812,7 @@ static void draw_plots(struct graph_info *ginfo, gint new_width)
 {
 	struct plot_list *list;
 	struct graph_plot *plot;
-	struct record *record;
+	struct pevent_record *record;
 	struct plot_hash *hash;
 	gint pid;
 	gint cpu;
@@ -1763,6 +1840,8 @@ static void draw_plots(struct graph_info *ginfo, gint new_width)
 	tracecmd_set_all_cpus_to_timestamp(ginfo->handle,
 					   ginfo->view_start_time);
 
+	trace_set_cursor(GDK_WATCH);
+
 	/* Shortcut if we don't have any task plots */
 	if (!ginfo->nr_task_hash && !ginfo->all_recs) {
 		for (cpu = 0; cpu < ginfo->cpus; cpu++) {
@@ -1817,9 +1896,11 @@ out:
 		plot = ginfo->plot_array[i];
 		draw_plot(ginfo, plot, NULL);
 		trace_graph_plot_end(ginfo, plot);
-		gdk_gc_unref(plot->gc);
+		if (plot->gc)
+			gdk_gc_unref(plot->gc);
 		plot->gc = NULL;
 	}
+	trace_put_cursor();
 }
 
 
@@ -1870,6 +1951,7 @@ static void draw_timeline(struct graph_info *ginfo, gint width)
 	gdk_draw_layout(ginfo->curr_pixmap, ginfo->draw->style->black_gc,
 			1, height+10, layout);
 	g_object_unref(layout);
+	trace_seq_destroy(&s);
 
 
 	/* --- draw ending time --- */
@@ -1883,6 +1965,7 @@ static void draw_timeline(struct graph_info *ginfo, gint width)
 	gdk_draw_layout(ginfo->curr_pixmap, ginfo->draw->style->black_gc,
 			width - (w + 2), height+10, layout);
 	g_object_unref(layout);
+	trace_seq_destroy(&s);
 
 
 	/* --- draw time at intervals --- */
@@ -1905,6 +1988,7 @@ static void draw_timeline(struct graph_info *ginfo, gint width)
 		gdk_draw_layout(ginfo->curr_pixmap, ginfo->draw->style->black_gc,
 				mid - (w / 2), height+10, layout);
 		g_object_unref(layout);
+		trace_seq_destroy(&s);
 	}
 }
 
@@ -2310,7 +2394,7 @@ void trace_graph_free_info(struct graph_info *ginfo)
 static int load_handle(struct graph_info *ginfo,
 		       struct tracecmd_input *handle)
 {
-	struct record *record;
+	struct pevent_record *record;
 	unsigned long sec, usec;
 	gint cpu;
 
@@ -2348,6 +2432,8 @@ static int load_handle(struct graph_info *ginfo,
 
 		free_record(record);
 		record = tracecmd_read_cpu_last(handle, cpu);
+		if (!record)
+			continue;
 
 		if (record->ts > ginfo->end_time)
 			ginfo->end_time = record->ts;
@@ -2395,6 +2481,132 @@ int trace_graph_load_handle(struct graph_info *ginfo,
 	return 0;
 }
 
+static int load_event_filter(struct graph_info *ginfo,
+			     struct tracecmd_xml_handle *handle,
+			     struct tracecmd_xml_system_node *node)
+{
+	struct tracecmd_xml_system_node *child;
+	struct event_filter *event_filter;
+	const char *name;
+	const char *value;
+
+	event_filter = ginfo->event_filter;
+
+	child = tracecmd_xml_node_child(node);
+	name = tracecmd_xml_node_type(child);
+	if (strcmp(name, "FilterType") != 0)
+		return -1;
+
+	value = tracecmd_xml_node_value(handle, child);
+	/* Do nothing with all events enabled */
+	if (strcmp(value, "all events") == 0)
+		return 0;
+
+	node = tracecmd_xml_node_next(child);
+	if (!node)
+		return -1;
+
+	pevent_filter_clear_trivial(event_filter, FILTER_TRIVIAL_BOTH);
+	ginfo->all_events = FALSE;
+
+	trace_filter_load_events(event_filter, handle, node);
+
+	return 0;
+}
+
+int trace_graph_load_filters(struct graph_info *ginfo,
+			     struct tracecmd_xml_handle *handle)
+{
+	struct tracecmd_xml_system *system;
+	struct tracecmd_xml_system_node *syschild;
+	const char *name;
+
+	if (filter_task_count(ginfo->task_filter) ||
+	    filter_task_count(ginfo->hide_tasks))
+		ginfo->filter_available = 1;
+	else
+		ginfo->filter_available = 0;
+
+	system = tracecmd_xml_find_system(handle, "TraceGraph");
+	if (!system)
+		return -1;
+
+	syschild = tracecmd_xml_system_node(system);
+	if (!syschild)
+		goto out_free_sys;
+
+	do {
+		name = tracecmd_xml_node_type(syschild);
+
+		if (strcmp(name, "EventFilter") == 0)
+			load_event_filter(ginfo, handle, syschild);
+
+		syschild = tracecmd_xml_node_next(syschild);
+	} while (syschild);
+
+	if (filter_task_count(ginfo->task_filter) ||
+	    filter_task_count(ginfo->hide_tasks))
+		ginfo->filter_available = 1;
+	else
+		ginfo->filter_available = 0;
+
+	tracecmd_xml_free_system(system);
+
+	trace_graph_refresh(ginfo);
+
+	return 0;
+
+ out_free_sys:
+	tracecmd_xml_free_system(system);
+	if (ginfo->filter_enabled)
+		trace_graph_refresh(ginfo);
+
+	return -1;
+}
+
+int trace_graph_save_filters(struct graph_info *ginfo,
+			     struct tracecmd_xml_handle *handle)
+{
+	struct event_filter *event_filter;
+
+	tracecmd_xml_start_system(handle, "TraceGraph");
+
+	event_filter = ginfo->event_filter;
+
+	tracecmd_xml_start_sub_system(handle, "EventFilter");
+
+	if (ginfo->all_events || !event_filter)
+		tracecmd_xml_write_element(handle, "FilterType", "all events");
+	else {
+		tracecmd_xml_write_element(handle, "FilterType", "filter");
+		trace_filter_save_events(handle, event_filter);
+	}
+
+	tracecmd_xml_end_sub_system(handle);
+
+	tracecmd_xml_end_system(handle);
+
+	return 0;
+}
+
+static void set_label_a(GtkWidget *widget)
+{
+	gtk_widget_set_tooltip_text(widget, "Click left mouse on graph\n"
+				    "to set Marker A");
+}
+
+static void set_label_b(GtkWidget *widget)
+{
+	gtk_widget_set_tooltip_text(widget, "Shift and click left mouse on graph\n"
+				    "to set Marker B");
+}
+
+static void set_label_cursor(GtkWidget *widget)
+{
+	gtk_widget_set_tooltip_text(widget, "Double click Left mouse on graph\n"
+				    "to set Cursor");
+}
+
 struct graph_info *
 trace_graph_create_with_callbacks(struct tracecmd_input *handle,
 				  struct graph_callbacks *cbs)
@@ -2454,11 +2666,13 @@ trace_graph_create_with_callbacks(struct tracecmd_input *handle,
 	/* --- Cursor --- */
 
 	label = gtk_label_new("Cursor:");
+	set_label_cursor(label);
 	gtk_table_attach(GTK_TABLE(table), label, 4, 5, 0, 1, GTK_EXPAND, GTK_EXPAND, 3, 3);
 	gtk_widget_show(label);
 
 	ginfo->cursor_label = gtk_label_new("0.0");
 	eventbox = gtk_event_box_new();
+	set_label_cursor(eventbox);
 	gtk_widget_show(eventbox);
 	gtk_widget_modify_bg(eventbox, GTK_STATE_NORMAL, &color);
 	gtk_container_add(GTK_CONTAINER(eventbox), ginfo->cursor_label);
@@ -2473,9 +2687,11 @@ trace_graph_create_with_callbacks(struct tracecmd_input *handle,
 	gtk_widget_show(hbox);
 
 	label = gtk_label_new("Marker");
+	set_label_a(label);
 	gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 0);
 	gtk_widget_show(label);
 
+
 	label = gtk_label_new("A:");
 
 	colorAB.red = 0;
@@ -2483,6 +2699,7 @@ trace_graph_create_with_callbacks(struct tracecmd_input *handle,
 	colorAB.blue = 0;
 
 	eventbox = gtk_event_box_new();
+	set_label_a(eventbox);
 	gtk_widget_show(eventbox);
 	gtk_widget_modify_bg(eventbox, GTK_STATE_NORMAL, &colorAB);
 	gtk_container_add(GTK_CONTAINER(eventbox), label);
@@ -2494,6 +2711,7 @@ trace_graph_create_with_callbacks(struct tracecmd_input *handle,
 
 	ginfo->marka_label = gtk_label_new("0.0");
 	eventbox = gtk_event_box_new();
+	set_label_a(eventbox);
 	gtk_widget_show(eventbox);
 	gtk_widget_modify_bg(eventbox, GTK_STATE_NORMAL, &color);
 	gtk_container_add(GTK_CONTAINER(eventbox), ginfo->marka_label);
@@ -2508,6 +2726,7 @@ trace_graph_create_with_callbacks(struct tracecmd_input *handle,
 	gtk_widget_show(hbox);
 
 	label = gtk_label_new("Marker");
+	set_label_b(label);
 	gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 0);
 	gtk_widget_show(label);
 
@@ -2518,6 +2737,7 @@ trace_graph_create_with_callbacks(struct tracecmd_input *handle,
 	colorAB.blue = 0;
 
 	eventbox = gtk_event_box_new();
+	set_label_b(eventbox);
 	gtk_widget_show(eventbox);
 	gtk_widget_modify_bg(eventbox, GTK_STATE_NORMAL, &colorAB);
 	gtk_container_add(GTK_CONTAINER(eventbox), label);
@@ -2530,6 +2750,7 @@ trace_graph_create_with_callbacks(struct tracecmd_input *handle,
 
 	ginfo->markb_label = gtk_label_new("0.0");
 	eventbox = gtk_event_box_new();
+	set_label_b(eventbox);
 	gtk_widget_show(eventbox);
 	gtk_widget_modify_bg(eventbox, GTK_STATE_NORMAL, &color);
 	gtk_container_add(GTK_CONTAINER(eventbox), ginfo->markb_label);
