@@ -19,6 +19,7 @@
  */
 #define _LARGEFILE64_SOURCE
 #include <dirent.h>
+#include <stdbool.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
@@ -72,9 +73,15 @@ struct cpu_data {
 	int			cpu;
 };
 
+struct buffer_instance {
+	char			*name;
+	size_t			offset;
+};
+
 struct tracecmd_input {
 	struct pevent		*pevent;
 	struct plugin_list	*plugin_list;
+	struct tracecmd_input	*parent;
 	unsigned long		flags;
 	int			fd;
 	int			long_size;
@@ -82,9 +89,12 @@ struct tracecmd_input {
 	int			read_page;
 	int			cpus;
 	int			ref;
+	int			nr_buffers;	/* buffer instances */
+	bool			use_trace_clock;
 	struct cpu_data 	*cpu_data;
 	unsigned long long	ts_offset;
 	char *			cpustats;
+	struct buffer_instance	*buffers;
 
 	struct tracecmd_ftrace	finfo;
 
@@ -1703,6 +1713,7 @@ static int handle_options(struct tracecmd_input *handle)
 	unsigned int size;
 	char *cpustats = NULL;
 	unsigned int cpustats_size = 0;
+	struct buffer_instance *buffer;
 	char *buf;
 
 	for (;;) {
@@ -1744,6 +1755,23 @@ static int handle_options(struct tracecmd_input *handle)
 			cpustats_size += size;
 			cpustats[cpustats_size] = 0;
 			break;
+		case TRACECMD_OPTION_BUFFER:
+			/* A buffer instance is saved at the end of the file */
+			handle->nr_buffers++;
+			handle->buffers = realloc(handle->buffers,
+						  sizeof(*handle->buffers) * handle->nr_buffers);
+			if (!handle->buffers)
+				die("realloc");
+			buffer = &handle->buffers[handle->nr_buffers - 1];
+			buffer->name = strdup(buf + 8);
+			if (!buffer->name)
+				die("strdup");
+			offset = *(unsigned long long *)buf;
+			buffer->offset = __data2host8(handle->pevent, offset);
+			break;
+		case TRACECMD_OPTION_TRACECLOCK:
+			handle->use_trace_clock = true;
+			break;
 		default:
 			warning("unknown option %d", option);
 			break;
@@ -1758,44 +1786,15 @@ static int handle_options(struct tracecmd_input *handle)
 	return 0;
 }
 
-/**
- * tracecmd_init_data - prepare reading the data from trace.dat
- * @handle: input handle for the trace.dat file
- *
- * This prepares reading the data from trace.dat. This is called
- * after tracecmd_read_headers() and before tracecmd_read_data().
- */
-int tracecmd_init_data(struct tracecmd_input *handle)
+static int read_cpu_data(struct tracecmd_input *handle)
 {
 	struct pevent *pevent = handle->pevent;
 	enum kbuffer_long_size long_size;
 	enum kbuffer_endian endian;
 	unsigned long long size;
-	char *cmdlines;
 	char buf[10];
 	int cpu;
 
-	size = read8(handle);
-	if (size < 0)
-		return -1;
-	cmdlines = malloc(size + 1);
-	if (!cmdlines)
-		return -1;
-	if (do_read_check(handle, cmdlines, size)) {
-		free(cmdlines);
-		return -1;
-	}
-	cmdlines[size] = 0;
-	parse_cmdlines(pevent, cmdlines, size);
-	free(cmdlines);
-
-	handle->cpus = read4(handle);
-	if (handle->cpus < 0)
-		return -1;
-
-	pevent_set_cpus(pevent, handle->cpus);
-	pevent_set_long_size(pevent, handle->long_size);
-
 	if (do_read_check(handle, buf, 10))
 		return -1;
 
@@ -1865,15 +1864,105 @@ int tracecmd_init_data(struct tracecmd_input *handle)
 			goto out_free;
 	}
 
-	tracecmd_blk_hack(handle);
-
 	return 0;
-out_free:
+
+ out_free:
 	for ( ; cpu >= 0; cpu--) {
 		free_page(handle, cpu);
 		kbuffer_free(handle->cpu_data[cpu].kbuf);
 	}
 	return -1;
+
+}
+
+static int read_data_and_size(struct tracecmd_input *handle,
+				     char **data, unsigned long long *size)
+{
+	*size = read8(handle);
+	if (*size < 0)
+		return -1;
+	*data = malloc(*size + 1);
+	if (!*data)
+		return -1;
+	if (do_read_check(handle, *data, *size)) {
+		free(*data);
+		return -1;
+	}
+
+	return 0;
+}
+
+static int read_and_parse_cmdlines(struct tracecmd_input *handle,
+							struct pevent *pevent)
+{
+	unsigned long long size;
+	char *cmdlines;
+
+	if (read_data_and_size(handle, &cmdlines, &size) < 0)
+		return -1;
+	cmdlines[size] = 0;
+	parse_cmdlines(pevent, cmdlines, size);
+	free(cmdlines);
+	return 0;
+}
+
+static int read_and_parse_trace_clock(struct tracecmd_input *handle,
+							struct pevent *pevent)
+{
+	unsigned long long size;
+	char *trace_clock;
+
+	if (read_data_and_size(handle, &trace_clock, &size) < 0)
+		return -1;
+	trace_clock[size] = 0;
+	parse_trace_clock(pevent, trace_clock, size);
+	free(trace_clock);
+	return 0;
+}
+
+/**
+ * tracecmd_init_data - prepare reading the data from trace.dat
+ * @handle: input handle for the trace.dat file
+ *
+ * This prepares reading the data from trace.dat. This is called
+ * after tracecmd_read_headers() and before tracecmd_read_data().
+ */
+int tracecmd_init_data(struct tracecmd_input *handle)
+{
+	struct pevent *pevent = handle->pevent;
+	int ret;
+
+	if (read_and_parse_cmdlines(handle, pevent) < 0)
+		return -1;
+
+	handle->cpus = read4(handle);
+	if (handle->cpus < 0)
+		return -1;
+
+	pevent_set_cpus(pevent, handle->cpus);
+	pevent_set_long_size(pevent, handle->long_size);
+
+	ret = read_cpu_data(handle);
+	if (ret < 0)
+		return ret;
+
+	if (handle->use_trace_clock) {
+		/*
+		 * There was a bug in the original setting of
+		 * the trace_clock file which let it get
+		 * corrupted. If it fails to read, force local
+		 * clock.
+		 */
+		if (read_and_parse_trace_clock(handle, pevent) < 0) {
+			char clock[] = "[local]";
+			warning("File has trace_clock bug, using local clock");
+			parse_trace_clock(pevent, clock, 8);
+		}
+	}
+
+	tracecmd_blk_hack(handle);
+
+	return ret;
 }
 
 /**
@@ -2130,10 +2219,15 @@ void tracecmd_close(struct tracecmd_input *handle)
 	}
 
 	free(handle->cpu_data);
-
 	close(handle->fd);
-	pevent_free(handle->pevent);
-	tracecmd_unload_plugins(handle->plugin_list);
+
+	if (handle->flags & TRACECMD_FL_BUFFER_INSTANCE)
+		tracecmd_close(handle->parent);
+	else {
+		/* Only main handle frees plugins and pevent */
+		pevent_free(handle->pevent);
+		tracecmd_unload_plugins(handle->plugin_list);
+	}
 	free(handle);
 }
 
@@ -2369,6 +2463,104 @@ int tracecmd_copy_headers(struct tracecmd_input *handle, int fd)
 	return 0;
 }
 
+/**
+ * tracecmd_record_at_buffer_start - return true if record is first on subbuffer
+ * @handle: input handle for the trace.dat file
+ * @record: The record to test if it is the first record on page
+ *
+ * Returns true if the record is the first record on the page.
+ */
+int tracecmd_record_at_buffer_start(struct tracecmd_input *handle,
+				    struct pevent_record *record)
+{
+	struct page *page = record->priv;
+	struct kbuffer *kbuf = handle->cpu_data[0].kbuf;
+	int offset;
+
+	if (!page || !kbuf)
+		return 0;
+
+	offset = record->offset - page->offset;
+	return offset == kbuffer_start_of_data(kbuf);
+}
+
+int tracecmd_buffer_instances(struct tracecmd_input *handle)
+{
+	return handle->nr_buffers;
+}
+
+const char *tracecmd_buffer_instance_name(struct tracecmd_input *handle, int indx)
+{
+	if (indx >= handle->nr_buffers)
+		return NULL;
+
+	return handle->buffers[indx].name;
+}
+
+struct tracecmd_input *
+tracecmd_buffer_instance_handle(struct tracecmd_input *handle, int indx)
+{
+	struct tracecmd_input *new_handle;
+	struct buffer_instance *buffer = &handle->buffers[indx];
+	size_t offset;
+	ssize_t ret;
+
+	if (indx >= handle->nr_buffers)
+		return NULL;
+
+	/*
+	 * We make a copy of the current handle, but we substitute
+	 * the cpu data with the cpu data for this buffer.
+	 */
+	new_handle = malloc(sizeof(*handle));
+	if (!new_handle)
+		return NULL;
+
+	*new_handle = *handle;
+	new_handle->cpu_data = NULL;
+	new_handle->nr_buffers = 0;
+	new_handle->buffers = NULL;
+	new_handle->ref = 1;
+	new_handle->parent = handle;
+	tracecmd_ref(handle);
+
+	new_handle->fd = dup(handle->fd);
+
+	new_handle->flags |= TRACECMD_FL_BUFFER_INSTANCE;
+
+	/* Save where we currently are */
+	offset = lseek64(handle->fd, 0, SEEK_CUR);
+
+	ret = lseek64(handle->fd, buffer->offset, SEEK_SET);
+	if (ret < 0) {
+		warning("could not seek to buffer %s offset %ld\n",
+			buffer->name, buffer->offset);
+		tracecmd_close(new_handle);
+		return NULL;
+	}
+
+	ret = read_cpu_data(new_handle);
+	if (ret < 0) {
+		warning("failed to read sub buffer %s\n", buffer->name);
+		tracecmd_close(new_handle);
+		return NULL;
+	}
+
+	ret = lseek64(handle->fd, offset, SEEK_SET);
+	if (ret < 0) {
+		warning("could not seek to back to offset %ld\n", offset);
+		tracecmd_close(new_handle);
+		return NULL;
+	}
+
+	return new_handle;
+}
+
+int tracecmd_is_buffer_instance(struct tracecmd_input *handle)
+{
+	return handle->flags & TRACECMD_FL_BUFFER_INSTANCE;
+}
+
 /**
  * tracecmd_long_size - return the size of "long" for the arch
  * @handle: input handle for the trace.dat file
@@ -2404,3 +2596,12 @@ struct pevent *tracecmd_get_pevent(struct tracecmd_input *handle)
 {
 	return handle->pevent;
 }
+
+/**
+ * tracecmd_get_use_trace_clock - return use_trace_clock
+ * @handle: input handle for the trace.dat file
+ */
+bool tracecmd_get_use_trace_clock(struct tracecmd_input *handle)
+{
+	return handle->use_trace_clock;
+}
