@@ -13,8 +13,7 @@
  * GNU General Public License for more details.
  *
  * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+ * along with this program; if not,  see <http://www.gnu.org/licenses>
  *
  * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  */
@@ -29,6 +28,7 @@
 struct task_plot_info {
 	int			pid;
 	struct cpu_data		*cpu_data;
+	struct pevent_record	**last_records;
 	unsigned long long	last_time;
 	unsigned long long	wake_time;
 	unsigned long long	display_wake_time;
@@ -54,10 +54,12 @@ static gint hash_pid(gint val)
 
 static int hash_cpu(int cpu)
 {
-	return trace_hash(cpu + 124);
+	cpu = (cpu << 3) + cpu * 21;
+ 
+	return trace_hash(cpu);
 }
 
-static gboolean is_running(struct graph_info *ginfo, struct record *record)
+static gboolean is_running(struct graph_info *ginfo, struct pevent_record *record)
 {
 	unsigned long long val;
 	int id;
@@ -71,7 +73,7 @@ static gboolean is_running(struct graph_info *ginfo, struct record *record)
 }
 
 static gboolean record_matches_pid(struct graph_info *ginfo,
-				   struct record *record, int match_pid,
+				   struct pevent_record *record, int match_pid,
 				   int *pid, int *sched_pid,
 				   gboolean *is_sched,
 				   gboolean *wakeup)
@@ -104,33 +106,38 @@ static gboolean record_matches_pid(struct graph_info *ginfo,
 	return FALSE;
 }
 
+static void set_cpu_to_time(int cpu, struct graph_info *ginfo, unsigned long long time)
+{
+	struct pevent_record *record;
+
+	tracecmd_set_cpu_to_timestamp(ginfo->handle, cpu, time);
+
+	while ((record = tracecmd_read_data(ginfo->handle, cpu))) {
+		if (record->ts >= time)
+			break;
+
+		free_record(record);
+	}
+	if (record) {
+		tracecmd_set_cursor(ginfo->handle, cpu, record->offset);
+		free_record(record);
+	} else
+		tracecmd_set_cpu_to_timestamp(ginfo->handle, cpu, time);
+}
+
 static void set_cpus_to_time(struct graph_info *ginfo, unsigned long long time)
 {
-	struct record *record;
 	int cpu;
 
-	for (cpu = 0; cpu < ginfo->cpus; cpu++) {
-		tracecmd_set_cpu_to_timestamp(ginfo->handle, cpu, time);
-
-		while ((record = tracecmd_read_data(ginfo->handle, cpu))) {
-			if (record->ts >= time)
-				break;
-
-			free_record(record);
-		}
-		if (record) {
-			tracecmd_set_cursor(ginfo->handle, cpu, record->offset);
-			free_record(record);
-		} else
-			tracecmd_set_cpu_to_timestamp(ginfo->handle, cpu, time);
-	}
+	for (cpu = 0; cpu < ginfo->cpus; cpu++)
+		set_cpu_to_time(cpu, ginfo, time);
 }
 
 static int task_plot_match_time(struct graph_info *ginfo, struct graph_plot *plot,
 			       unsigned long long time)
 {
 	struct task_plot_info *task_info = plot->private;
-	struct record *record = NULL;
+	struct pevent_record *record = NULL;
 	gboolean is_wakeup;
 	gboolean is_sched;
 	gboolean match;
@@ -172,7 +179,7 @@ struct offset_cache {
 static struct offset_cache *save_offsets(struct graph_info *ginfo)
 {
 	struct offset_cache *offsets;
-	struct record *record;
+	struct pevent_record *record;
 	int cpu;
 
 	offsets = malloc_or_die(sizeof(*offsets));
@@ -190,7 +197,7 @@ static struct offset_cache *save_offsets(struct graph_info *ginfo)
 
 static void restore_offsets(struct graph_info *ginfo, struct offset_cache *offsets)
 {
-	struct record *record;
+	struct pevent_record *record;
 	int cpu;
 
 	for (cpu = 0; cpu < ginfo->cpus; cpu++) {
@@ -207,10 +214,10 @@ static void restore_offsets(struct graph_info *ginfo, struct offset_cache *offse
 	free(offsets);
 }
 
-static struct record *
+static struct pevent_record *
 find_record(struct graph_info *ginfo, gint pid, guint64 time)
 {
-	struct record *record = NULL;
+	struct pevent_record *record = NULL;
 	gboolean is_wakeup;
 	gboolean is_sched;
 	gboolean match;
@@ -244,7 +251,7 @@ static int task_plot_display_last_event(struct graph_info *ginfo,
 {
 	struct task_plot_info *task_info = plot->private;
 	struct event_format *event;
-	struct record *record;
+	struct pevent_record *record;
 	struct offset_cache *offsets;
 	gboolean is_sched;
 	gboolean is_wakeup;
@@ -315,15 +322,98 @@ static void task_plot_start(struct graph_info *ginfo, struct graph_plot *plot,
 {
 	struct task_plot_info *task_info = plot->private;
 
+	memset(task_info->last_records, 0, sizeof(struct pevent_record *) * ginfo->cpus);
+
 	task_info->last_time = 0ULL;
 	task_info->last_cpu = -1;
 	task_info->wake_time = 0ULL;
 	task_info->display_wake_time = 0ULL;
 }
 
+static void update_last_record(struct graph_info *ginfo,
+			       struct task_plot_info *task_info,
+			       struct pevent_record *record)
+{
+	struct tracecmd_input *handle = ginfo->handle;
+	struct pevent_record *trecord, *t2record;
+	struct pevent_record *saved;
+	unsigned long long ts;
+	int sched_pid;
+	int pid;
+	int rec_pid;
+	int is_wakeup;
+	int is_sched;
+	int this_cpu;
+	int cpu;
+
+	pid = task_info->pid;
+
+	if (record) {
+		ts = record->ts;
+		this_cpu = record->cpu;
+	} else {
+		ts = ginfo->view_end_time;
+		this_cpu = -1;
+	}
+
+	for (cpu = 0; cpu < ginfo->cpus; cpu++) {
+
+		if (task_info->last_records[cpu])
+			continue;
+
+		if (cpu == this_cpu) {
+			static int once;
+
+			trecord = tracecmd_read_prev(handle, record);
+			/* Set cpu cursor back to what it was  */
+			saved = tracecmd_read_data(handle, cpu);
+			if (!once && saved->offset != record->offset) {
+				once++;
+				warning("failed to reset cursor!");
+			}
+			free_record(saved);
+		} else {
+			static int once;
+
+			saved = tracecmd_read_data(handle, cpu);
+			set_cpu_to_time(cpu, ginfo, ts);
+			t2record = tracecmd_read_data(handle, cpu);
+			trecord = tracecmd_read_prev(handle, t2record);
+			free_record(t2record);
+			/* reset cursor back to what it was */
+			if (saved) {
+				tracecmd_set_cursor(handle, cpu, saved->offset);
+				free_record(saved);
+			} else {
+				saved = tracecmd_read_data(handle, cpu);
+				if (!once && saved) {
+					once++;
+					warning("failed to reset cursor to end!");
+				}
+				/* saved should always be NULL */
+				free_record(saved);
+			}
+		}
+		if (!trecord)
+			continue;
+
+		if (record_matches_pid(ginfo, trecord, pid, &rec_pid,
+				       &sched_pid, &is_sched, &is_wakeup) &&
+		    !is_wakeup &&
+		    (!is_sched || (is_sched && sched_pid == pid))) {
+			task_info->last_records[cpu] = trecord;
+			task_info->last_cpu = trecord->cpu;
+			task_info->last_time = trecord->ts;
+			break;
+		}
+
+		free_record(trecord);
+	}
+}
+
 static int task_plot_event(struct graph_info *ginfo,
 			   struct graph_plot *plot,
-			   struct record *record,
+			   struct pevent_record *record,
 			   struct plot_info *info)
 {
 	struct task_plot_info *task_info = plot->private;
@@ -333,10 +423,12 @@ static int task_plot_event(struct graph_info *ginfo,
 	int is_wakeup;
 	int is_sched;
 	int pid;
+	int cpu;
 
 	pid = task_info->pid;
 
 	if (!record) {
+		update_last_record(ginfo, task_info, record);
 		/* no more records, finish a box if one was started */
 		if (task_info->last_cpu >= 0) {
 			info->box = TRUE;
@@ -344,20 +436,39 @@ static int task_plot_event(struct graph_info *ginfo,
 			info->bend = ginfo->view_end_time;
 			info->bcolor = hash_cpu(task_info->last_cpu);
 		}
+		for (cpu = 0; cpu < ginfo->cpus; cpu++) {
+			free_record(task_info->last_records[cpu]);
+			task_info->last_records[cpu] = NULL;
+		}
 		return 0;
 	}
 
 	match = record_matches_pid(ginfo, record, pid, &rec_pid,
 				   &sched_pid, &is_sched, &is_wakeup);
 
-	if (!match && record->cpu != task_info->last_cpu)
+
+	if (!match && record->cpu != task_info->last_cpu) {
+		if (!task_info->last_records[record->cpu]) {
+			task_info->last_records[record->cpu] = record;
+			tracecmd_record_ref(record);
+		}
 		return 0;
+	}
 
 	if (match) {
 		info->line = TRUE;
 		info->lcolor = hash_pid(rec_pid);
 		info->ltime = record->ts;
 
+		/*
+		 * Is this our first match?
+		 *
+		 * If last record is NULL, then it may exist off the
+		 * viewable range. Search to see if one exists, and if
+		 * it is the record we want to match.
+		 */
+		update_last_record(ginfo, task_info, record);
+
 		if (is_wakeup) {
 			/* Wake up but not task */
 			info->ltime = hash_pid(rec_pid);
@@ -424,6 +535,12 @@ static int task_plot_event(struct graph_info *ginfo,
 		return 1;
 	}
 
+	cpu = record->cpu;
+
+	if (!task_info->last_records[cpu]) {
+		task_info->last_records[cpu] = record;
+		tracecmd_record_ref(record);
+	}
 	/* not a match, and on the last CPU, scheduled out? */
 	if (task_info->last_cpu >= 0) {
 		info->box = TRUE;
@@ -437,7 +554,7 @@ static int task_plot_event(struct graph_info *ginfo,
 }
 
 
-static struct record *
+static struct pevent_record *
 task_plot_find_record(struct graph_info *ginfo, struct graph_plot *plot,
 		      unsigned long long time)
 {
@@ -451,12 +568,12 @@ task_plot_find_record(struct graph_info *ginfo, struct graph_plot *plot,
 
 #define MAX_SEARCH 20
 
-static struct record *
-find_previous_record(struct graph_info *ginfo, struct record *start_record,
+static struct pevent_record *
+find_previous_record(struct graph_info *ginfo, struct pevent_record *start_record,
 		     int pid, int cpu)
 {
-	struct record *last_record = start_record;
-	struct record *record;
+	struct pevent_record *last_record = start_record;
+	struct pevent_record *record;
 	gboolean match;
 	gboolean is_sched;
 	gboolean is_wakeup;
@@ -464,7 +581,9 @@ find_previous_record(struct graph_info *ginfo, struct record *start_record,
 	gint sched_pid;
 	int count = 0;
 
-	if (!last_record)
+	if (last_record)
+		last_record->ref_count++;
+	else
 		last_record = tracecmd_read_cpu_last(ginfo->handle, cpu);
 
 	while ((record = tracecmd_read_prev(ginfo->handle, last_record))) {
@@ -475,8 +594,7 @@ find_previous_record(struct graph_info *ginfo, struct record *start_record,
 		if (match)
 			break;
 
-		if (last_record != start_record)
-			free_record(last_record);
+		free_record(last_record);
 
 		if (count > MAX_SEARCH) {
 			free_record(record);
@@ -485,17 +603,16 @@ find_previous_record(struct graph_info *ginfo, struct record *start_record,
 		last_record = record;
 	}
 
-	if (last_record != start_record)
-		free_record(last_record);
+	free_record(last_record);
 
 	return record;
 }
 
-static struct record *
+static struct pevent_record *
 get_display_record(struct graph_info *ginfo, int pid, unsigned long long time)
 {
-	struct record *record;
-	struct record **records;
+	struct pevent_record *record;
+	struct pevent_record **records;
 	unsigned long long ts;
 	int next_cpu;
 	int cpu;
@@ -567,11 +684,13 @@ int task_plot_display_info(struct graph_info *ginfo,
 {
 	struct task_plot_info *task_info = plot->private;
 	struct event_format *event;
-	struct record *record;
+	struct pevent_record *record;
 	struct pevent *pevent;
 	unsigned long sec, usec;
 	const char *comm;
+	int cpu;
 	int type;
+	int sched_pid = -1;
 	int pid;
 
 	pid = task_info->pid;
@@ -582,7 +701,7 @@ int task_plot_display_info(struct graph_info *ginfo,
 	pevent = ginfo->pevent;
 
 	pid = pevent_data_pid(ginfo->pevent, record);
-	comm = pevent_data_comm_from_pid(ginfo->pevent, pid);
+	cpu = record->cpu;
 
 	convert_nano(record->ts, &sec, &usec);
 
@@ -600,16 +719,12 @@ int task_plot_display_info(struct graph_info *ginfo,
 			trace_seq_putc(s, '\n');
 		} else
 			trace_seq_printf(s, "UNKNOW EVENT %d\n", type);
-	} else {
-		if (record->ts < time)
-			trace_graph_check_sched_switch(ginfo, record, &pid, &comm);
 	}
+	trace_graph_check_sched_switch(ginfo, record, &sched_pid, &comm);
 
 	trace_seq_printf(s, "%lu.%06lu", sec, usec);
-	if (pid)
-		trace_seq_printf(s, " %s-%d", comm, pid);
-	else
-		trace_seq_puts(s, " <idle>");
+	if (pid == task_info->pid || sched_pid == task_info->pid)
+		trace_seq_printf(s, " CPU: %03d", cpu);
 
 	free_record(record);
 
@@ -622,6 +737,7 @@ void task_plot_destroy(struct graph_info *ginfo, struct graph_plot *plot)
 
 	trace_graph_plot_remove_all_recs(ginfo, plot);
 
+	free(task_info->last_records);
 	free(task_info);
 }
 
@@ -726,7 +842,7 @@ void graph_plot_init_tasks(struct graph_info *ginfo)
 {
 	struct task_plot_info *task_info;
 	char label[100];
-	struct record *record;
+	struct pevent_record *record;
 	int pid;
 
 	/* Just for testing */
@@ -740,6 +856,8 @@ void graph_plot_init_tasks(struct graph_info *ginfo)
 	}
 
 	task_info = malloc_or_die(sizeof(*task_info));
+	task_info->last_records =
+		malloc_or_die(sizeof(struct pevent_record *) * ginfo->cpus);
 	task_info->pid = pid;
 
 	snprintf(label, 100, "TASK %d", pid);
@@ -756,6 +874,8 @@ void graph_plot_task(struct graph_info *ginfo, int pid, int pos)
 	int len;
 
 	task_info = malloc_or_die(sizeof(*task_info));
+	task_info->last_records =
+		malloc_or_die(sizeof(struct pevent_record *) * ginfo->cpus);
 	task_info->pid = pid;
 	comm = pevent_data_comm_from_pid(ginfo->pevent, pid);
 
