@@ -16,8 +16,7 @@
  * GNU General Public License for more details.
  *
  * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+ * along with this program; if not,  see <http://www.gnu.org/licenses>
  *
  * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  */
@@ -387,18 +386,18 @@ trace_view_store_get_path (GtkTreeModel *tree_model,
 {
 	GtkTreePath	*path;
 	TraceViewRecord *record;
-	TraceViewStore	*trace_view_store;
+	TraceViewStore	*store;
 
 	g_return_val_if_fail (TRACE_VIEW_IS_LIST(tree_model), NULL);
 	g_return_val_if_fail (iter != NULL,	NULL);
 	g_return_val_if_fail (iter->user_data != NULL,	NULL);
 
-	trace_view_store = TRACE_VIEW_STORE(tree_model);
+	store = TRACE_VIEW_STORE(tree_model);
 
 	record = (TraceViewRecord*) iter->user_data;
 
 	path = gtk_tree_path_new();
-	gtk_tree_path_append_index(path, record->pos);
+	gtk_tree_path_append_index(path, record->pos - store->start_row);
 
 	return path;
 }
@@ -422,11 +421,11 @@ trace_view_store_get_value (GtkTreeModel *tree_model,
 	struct trace_seq s;
 	struct pevent *pevent;
 	struct event_format *event;
-	struct record *data;
+	struct pevent_record *data;
 	const gchar *comm;
 	gchar *str;
 	guint64 secs, usecs;
-	gint val, pos;
+	gint val;
 	int cpu;
 
 	g_return_if_fail (TRACE_VIEW_IS_LIST (tree_model));
@@ -443,11 +442,6 @@ trace_view_store_get_value (GtkTreeModel *tree_model,
 
 	g_return_if_fail ( record != NULL );
 
-	pos = record->pos - trace_view_store->start_row;
-
-	if(pos >= trace_view_store->num_rows)
-		g_return_if_reached();
-
 	column = get_visible_column(TRACE_VIEW_STORE(tree_model), column);
 
 	switch(column)
@@ -500,21 +494,28 @@ trace_view_store_get_value (GtkTreeModel *tree_model,
 			trace_seq_init(&s);
 			pevent_data_lat_fmt(pevent, &s, data);
 			g_value_set_string(value, s.buffer);
+			trace_seq_destroy(&s);
 			break;
 
 		case TRACE_VIEW_STORE_COL_EVENT:
 		case TRACE_VIEW_STORE_COL_INFO:
 			val = pevent_data_type(pevent, data);
 			event = pevent_data_event_from_type(pevent, val);
+			if (!event) {
+				if (column == TRACE_VIEW_STORE_COL_EVENT)
+					g_value_set_string(value, "[UNKNOWN EVENT]");
+				break;
+			}
+
 			if (column == TRACE_VIEW_STORE_COL_EVENT) {
 				g_value_set_string(value, event->name);
 				break;
 			}
 
-			
 			trace_seq_init(&s);
 			pevent_event_info(&s, event, data);
 			g_value_set_string(value, s.buffer);
+			trace_seq_destroy(&s);
 			break;
 		}
 		free_record(data);
@@ -853,7 +854,7 @@ TraceViewStore *
 trace_view_store_new (struct tracecmd_input *handle)
 {
 	TraceViewStore *newstore;
-	struct record *data;
+	struct pevent_record *data;
 	gint cpu, count, total=0;
 	struct temp {
 		guint64		offset;
@@ -1099,8 +1100,9 @@ gint trace_view_store_get_timestamp_page(TraceViewStore *store, guint64 ts)
 static TraceViewRecord *get_row(TraceViewStore *store, gint row)
 {
 	TraceViewRecord *record;
+	gint index = row - store->start_row;
 
-	g_return_val_if_fail(row >= store->start_row && row < store->visible_rows, NULL);
+	g_return_val_if_fail(index >= 0 && index < store->visible_rows, NULL);
 
 	record = store->rows[row];
 	g_assert(record != NULL);
@@ -1132,7 +1134,26 @@ trace_view_store_get_visible_row(TraceViewStore *store, gint row)
 	return get_row(store, row);
 }
 
-gint get_next_pid(TraceViewStore *store, struct pevent *pevent, struct record *record)
+TraceViewRecord *
+trace_view_store_get_actual_row(TraceViewStore *store, gint row)
+{
+	g_return_val_if_fail (TRACE_VIEW_IS_LIST (store), NULL);
+	g_return_val_if_fail (row >= 0, NULL);
+	g_return_val_if_fail (row < store->actual_rows, NULL);
+
+	if (!store->rows)
+		return NULL;
+
+	return store->rows[row];
+}
+
+gint trace_view_store_get_num_actual_rows(TraceViewStore *store)
+{
+	g_return_val_if_fail (TRACE_VIEW_IS_LIST (store), -1);
+	return store->actual_rows;
+}
+
+gint get_next_pid(TraceViewStore *store, struct pevent *pevent, struct pevent_record *record)
 {
 	unsigned long long val;
 	int ret;
@@ -1142,7 +1163,7 @@ gint get_next_pid(TraceViewStore *store, struct pevent *pevent, struct record *r
 	return val;
 }
 
-gint get_wakeup_pid(TraceViewStore *store, struct pevent *pevent, struct record *record)
+gint get_wakeup_pid(TraceViewStore *store, struct pevent *pevent, struct pevent_record *record)
 {
 	unsigned long long val;
 	int ret;
@@ -1152,7 +1173,7 @@ gint get_wakeup_pid(TraceViewStore *store, struct pevent *pevent, struct record
 	return val;
 }
 
-gint get_wakeup_new_pid(TraceViewStore *store, struct pevent *pevent, struct record *record)
+gint get_wakeup_new_pid(TraceViewStore *store, struct pevent *pevent, struct pevent_record *record)
 {
 	unsigned long long val;
 	int ret;
@@ -1173,7 +1194,7 @@ static gboolean view_task(TraceViewStore *store, gint pid)
 }
 
 static gboolean show_task(TraceViewStore *store, struct pevent *pevent,
-			  struct record *record, gint pid)
+			  struct pevent_record *record, gint pid)
 {
 	gint event_id;
 
@@ -1213,7 +1234,7 @@ static void update_filter_tasks(TraceViewStore *store)
 {
 	struct tracecmd_input *handle;
 	struct pevent *pevent;
-	struct record *record;
+	struct pevent_record *record;
 	gint pid;
 	gint cpu;
 	gint i;
@@ -1330,10 +1351,8 @@ void trace_view_store_assign_filters(TraceViewStore *store,
 
 	if (store->task_filter != task_filter)
 		store->task_filter = filter_task_hash_copy(task_filter);
-
 }
 
-
 /*****************************************************************************
  *
  *	trace_view_store_append_record:	Empty lists are boring. This function can
