@@ -13,12 +13,10 @@
  * GNU General Public License for more details.
  *
  * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+ * along with this program; if not,  see <http://www.gnu.org/licenses>
  *
  * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  */
-#define _GNU_SOURCE
 #include <stdio.h>
 #include <string.h>
 #include <stdarg.h>
@@ -28,9 +26,10 @@
 #include "trace-cmd.h"
 #include "trace-local.h"
 #include "trace-view.h"
+#include "trace-gui.h"
 #include "trace-compat.h"
 #include "cpu.h"
-#include "util.h"
+#include "event-utils.h"
 
 enum {
 	COL_INDEX,
@@ -170,6 +169,16 @@ trace_view_load(GtkWidget *view, struct tracecmd_input *handle,
 	GtkTreeModel *model;
 	long c;
 
+	/*
+	 * The handle should have loaded the file by now.
+	 * Try to turn off function trace indent and turn on show parent
+	 * if possible.
+	 */
+	trace_util_add_option("ftrace:parent", "1");
+	trace_util_add_option("ftrace:indent", "0");
+
+	/* Also, show the function name in the tail for function graph */
+	trace_util_add_option("fgraph:tailprint", "1");
 
 	/* --- CPU column --- */
 
@@ -682,6 +691,87 @@ static gboolean test_text(const gchar *text, const gchar *search_text, enum sele
 	return FALSE;
 }
 
+static gboolean test_row(GtkTreeModel *model, GtkTreeIter *iter, gint sel,
+			 gint col_num, gint search_val, const gchar *search_text)
+{
+	gchar *text = NULL;
+	gboolean found = FALSE;
+	gint val;
+
+	switch (col_num) {
+	case TRACE_VIEW_STORE_COL_INDEX:
+	case TRACE_VIEW_STORE_COL_CPU:
+	case TRACE_VIEW_STORE_COL_PID:
+		/* integers */
+
+		gtk_tree_model_get(model, iter,
+				   col_num, &val,
+				   -1);
+		if (test_int(val, search_val, sel))
+			found = TRUE;
+		break;
+
+	case TRACE_VIEW_STORE_COL_TS:
+	case TRACE_VIEW_STORE_COL_COMM:
+	case TRACE_VIEW_STORE_COL_LAT:
+	case TRACE_VIEW_STORE_COL_EVENT:
+	case TRACE_VIEW_STORE_COL_INFO:
+		/* strings */
+
+		gtk_tree_model_get(model, iter,
+				   col_num, &text,
+				   -1);
+
+		if (test_text(text, search_text, sel))
+			found = TRUE;
+		break;
+	}
+
+	return found;
+}
+
+static void search_next_pages(GtkTreeView *tree,
+			      TraceViewStore *store, gint sel, gint col_num,
+			      gint search_val, const char *search_text)
+{
+	GtkTreeModel *model;
+	TraceViewRecord *rec;
+	GtkTreeIter iter;
+	gint row;
+	gint total_rows;
+	gboolean found = FALSE;
+
+	model = (GtkTreeModel *)store;
+
+	row = store->start_row + store->num_rows;
+	total_rows = trace_view_store_get_num_actual_rows(store);
+
+	trace_set_cursor(GDK_WATCH);
+	trace_freeze_all();
+
+	for (; row < total_rows; row++) {
+
+		/* Needed to process the cursor change */
+		if (!(row & ((1 << 5)-1)))
+			gtk_main_iteration_do(FALSE);
+
+		rec = trace_view_store_get_actual_row(store, row);
+		iter.user_data = rec;
+		found = test_row(model, &iter, sel, col_num, search_val, search_text);
+		if (found)
+			break;
+	}
+	trace_unfreeze_all();
+	trace_put_cursor();
+
+	if (!found) {
+		trace_dialog(NULL, TRACE_GUI_INFO, "Not found");
+		return;
+	}
+
+	trace_view_select(GTK_WIDGET(tree), rec->timestamp);
+}
+
 static void search_tree(gpointer data)
 {
 	struct search_info *info = data;
@@ -694,14 +784,13 @@ static void search_tree(gpointer data)
 	GtkComboBox *col_combo = GTK_COMBO_BOX(info->column);
 	GtkComboBox *sel_combo = GTK_COMBO_BOX(info->selection);
 	const gchar *title;
-	gint val;
-	gchar *text = NULL;
 	const gchar *search_text;
 	gint col_num;
 	gint sel;
 	gint search_val;
 	gint start_row;
 	gboolean found = FALSE;
+	gint i = 0;
 
 	col_num = gtk_combo_box_get_active(col_combo);
 	sel = gtk_combo_box_get_active(sel_combo);
@@ -737,45 +826,37 @@ static void search_tree(gpointer data)
 	if (!gtk_tree_model_iter_nth_child(model, &iter, NULL, start_row))
 		return;
 
+	trace_set_cursor(GDK_WATCH);
+	trace_freeze_all();
+
 	search_val = atoi(search_text);
 	while (gtk_tree_model_iter_next(model, &iter)) {
-		switch (col_num) {
-		case TRACE_VIEW_STORE_COL_INDEX:
-		case TRACE_VIEW_STORE_COL_CPU:
-		case TRACE_VIEW_STORE_COL_PID:
-		/* integers */
-
-			gtk_tree_model_get(model, &iter,
-					   col_num, &val,
-					   -1);
-			if (test_int(val, search_val, sel))
-				found = TRUE;
-			break;
-
-		case TRACE_VIEW_STORE_COL_TS:
-		case TRACE_VIEW_STORE_COL_COMM:
-		case TRACE_VIEW_STORE_COL_LAT:
-		case TRACE_VIEW_STORE_COL_EVENT:
-		case TRACE_VIEW_STORE_COL_INFO:
-			/* strings */
-
-			gtk_tree_model_get(model, &iter,
-					   col_num, &text,
-					   -1);
-
-			if (test_text(text, search_text, sel))
-				found = TRUE;
-			break;
-		}
+		/* Needed to process the cursor change */
+		if (!(i++ & ((1 << 5)-1)))
+		    gtk_main_iteration_do(FALSE);
 
+		found = test_row(model, &iter, sel, col_num, search_val, search_text);
 		if (found)
 			break;
 	}
-
+	trace_unfreeze_all();
+	trace_put_cursor();
 
 	if (!found) {
-		printf("NOT FOUND!\n");
-		/* show pop up */
+		GtkResponseType ret;
+		gint pages = trace_view_store_get_pages(store);
+		gint page = trace_view_store_get_page(store);
+
+		if (page < pages) {
+			ret = trace_dialog(NULL, TRACE_GUI_ASK,
+					   "Not found on this page\n"
+					   "Search next pages?");
+			if (ret == GTK_RESPONSE_YES)
+				search_next_pages(info->treeview, store, sel,
+						  col_num, search_val, search_text);
+			return;
+		}
+		trace_dialog(NULL, TRACE_GUI_INFO, "Not found");
 		return;
 	}
 
@@ -870,3 +951,112 @@ void trace_view_search_setup(GtkBox *box, GtkTreeView *treeview)
 				  G_CALLBACK (search_tree),
 				  (gpointer) info);
 }
+
+int trace_view_save_filters(struct tracecmd_xml_handle *handle,
+			    GtkTreeView *trace_tree)
+{
+	struct event_filter *event_filter;
+	GtkTreeModel *model;
+	TraceViewStore *store;
+	gboolean all_events;
+
+	model = gtk_tree_view_get_model(trace_tree);
+	if (!model)
+		return -1;
+
+	store = TRACE_VIEW_STORE(model);
+
+	tracecmd_xml_start_system(handle, "TraceView");
+
+	all_events = trace_view_store_get_all_events_enabled(store);
+	event_filter = trace_view_store_get_event_filter(store);
+
+	tracecmd_xml_start_sub_system(handle, "EventFilter");
+
+	if (all_events || !event_filter)
+		tracecmd_xml_write_element(handle, "FilterType", "all events");
+	else {
+		tracecmd_xml_write_element(handle, "FilterType", "filter");
+		trace_filter_save_events(handle, event_filter);
+	}
+
+	tracecmd_xml_end_sub_system(handle);
+
+	tracecmd_xml_end_system(handle);
+
+	return 0;
+}
+
+static int load_event_filter(TraceViewStore *store,
+			     struct tracecmd_xml_handle *handle,
+			     struct tracecmd_xml_system_node *node)
+{
+	struct tracecmd_xml_system_node *child;
+	struct event_filter *event_filter;
+	const char *name;
+	const char *value;
+
+	event_filter = trace_view_store_get_event_filter(store);
+
+	child = tracecmd_xml_node_child(node);
+	name = tracecmd_xml_node_type(child);
+	if (strcmp(name, "FilterType") != 0)
+		return -1;
+
+	value = tracecmd_xml_node_value(handle, child);
+	/* Do nothing with all events enabled */
+	if (strcmp(value, "all events") == 0)
+		return 0;
+
+	node = tracecmd_xml_node_next(child);
+	if (!node)
+		return -1;
+
+	trace_view_store_clear_all_events_enabled(store);
+
+	trace_filter_load_events(event_filter, handle, node);
+
+	return 0;
+}
+
+int trace_view_load_filters(struct tracecmd_xml_handle *handle,
+			    GtkTreeView *trace_tree)
+{
+	struct tracecmd_xml_system *system;
+	struct tracecmd_xml_system_node *syschild;
+	GtkTreeModel *model;
+	TraceViewStore *store;
+	const char *name;
+
+	model = gtk_tree_view_get_model(trace_tree);
+	if (!model)
+		return -1;
+
+	store = TRACE_VIEW_STORE(model);
+
+	system = tracecmd_xml_find_system(handle, "TraceView");
+	if (!system)
+		return -1;
+
+	syschild = tracecmd_xml_system_node(system);
+	if (!syschild)
+		goto out_free_sys;
+
+	do {
+		name = tracecmd_xml_node_type(syschild);
+
+		if (strcmp(name, "EventFilter") == 0)
+			load_event_filter(store, handle, syschild);
+
+		syschild = tracecmd_xml_node_next(syschild);
+	} while (syschild);
+
+	tracecmd_xml_free_system(system);
+
+	update_rows(trace_tree, store);
+	return 0;
+
+ out_free_sys:
+	tracecmd_xml_free_system(system);
+	return -1;
+}
